#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: DeadCanadian
# 	started naming
# revision: 3		author: Lord Zedd
# 	I don't always add revisions, but when I do it's because I spend hours making these giant enums.
# revision: 4		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef

sily_parameter = (
    "int_global_teams",
    "int_global_round_reset",
    "int_global_round_time_limit",
    "int_global_round_count",
    "int_global_early_victory_win_count",
    "int_global_respawn_inheritance",
    "int_global_respawn_with_team",
    "int_global_respawn_at_location",
    "int_global_respawn_on_kill",
    "int_global_lives_per_round",
    "int_global_team_lives_per_round",
    "int_global_respawn_time",
    "int_global_suicide_penalty",
    "int_global_betrayal_penalty",
    "int_global_respawn_growth",
    "int_global_respawn_traits_duration",
    "int_global_respawn_traits_damage_resistance",
    "int_global_respawn_traits_shield_multiplier",
    "int_global_respawn_traits_shield_recharge",
    "int_global_respawn_traits_headshot_immunity",
    "int_global_respawn_traits_vampirism",
    "int_global_respawn_traits_damage_modifier",
    "sid_global_respawn_traits_primary_weapon",
    "sid_global_respawn_traits_secondary_weapon",
    "int_global_respawn_traits_grenade_count",
    "int_global_respawn_traits_grenade_recharge",
    "int_global_respawn_traits_infinite_ammo",
    "int_global_respawn_traits_weapon_pickup",
    "int_global_respawn_traits_speed",
    "int_global_respawn_traits_gravity",
    "int_global_respawn_traits_vehicle_use",
    "int_global_respawn_traits_sensor",
    "int_global_respawn_traits_sensor_range",
    "int_global_respawn_traits_waypoint",
    "int_global_respawn_traits_active_camo",
    "int_global_respawn_traits_visual_aura",
    "int_global_respawn_traits_forced_color",
    "int_global_observers",
    "int_global_team_changing",
    "int_global_friendly_fire",
    "int_global_betrayal_booting",
    "int_global_enemy_voice",
    "int_global_open_channel_voice",
    "int_global_dead_player_voice",
    "int_global_grenades_on_map",
    "int_global_indestructible_vehicles",
    "int_global_base_traits_damage_resistance",
    "int_global_base_traits_shield_multiplier",
    "int_global_base_traits_shield_recharge",
    "int_global_base_traits_headshot_immunity",
    "int_global_base_traits_vampirism",
    "int_global_base_traits_damage_modifier",
    "sid_global_base_traits_primary_weapon",
    "sid_global_base_traits_secondary_weapon",
    "int_global_base_traits_grenade_count",
    "int_global_base_traits_grenade_recharge",
    "int_global_base_traits_infinite_ammo",
    "int_global_base_traits_weapon_pickup",
    "int_global_base_traits_speed",
    "int_global_base_traits_gravity",
    "int_global_base_traits_vehicle_use",
    "int_global_base_traits_sensor",
    "int_global_base_traits_sensor_range",
    "int_global_base_traits_waypoint",
    "int_global_base_traits_active_camo",
    "int_global_base_traits_visual_aura",
    "int_global_base_traits_forced_color",
    "sid_global_map_weapon_set",
    "sid_global_map_vehicle_set",
    "int_global_red_traits_damage_resistance",
    "int_global_red_traits_shield_multiplier",
    "int_global_red_traits_shield_recharge",
    "int_global_red_traits_headshot_immunity",
    "int_global_red_traits_vampirism",
    "int_global_red_traits_damage_modifier",
    "sid_global_red_traits_primary_weapon",
    "sid_global_red_traits_secondary_weapon",
    "int_global_red_traits_grenade_count",
    "int_global_red_traits_grenade_recharge",
    "int_global_red_traits_infinite_ammo",
    "int_global_red_traits_weapon_pickup",
    "int_global_red_traits_speed",
    "int_global_red_traits_gravity",
    "int_global_red_traits_vehicle_use",
    "int_global_red_traits_sensor",
    "int_global_red_traits_sensor_range",
    "int_global_red_traits_waypoint",
    "int_global_red_traits_active_camo",
    "int_global_red_traits_visual_aura",
    "int_global_red_traits_forced_color",
    "int_global_blue_traits_damage_resistance",
    "int_global_blue_traits_shield_multiplier",
    "int_global_blue_traits_shield_recharge",
    "int_global_blue_traits_headshot_immunity",
    "int_global_blue_traits_vampirism",
    "int_global_blue_traits_damage_modifier",
    "sid_global_blue_traits_primary_weapon",
    "sid_global_blue_traits_secondary_weapon",
    "int_global_blue_traits_grenade_count",
    "int_global_blue_traits_grenade_recharge",
    "int_global_blue_traits_infinite_ammo",
    "int_global_blue_traits_weapon_pickup",
    "int_global_blue_traits_speed",
    "int_global_blue_traits_gravity",
    "int_global_blue_traits_vehicle_use",
    "int_global_blue_traits_sensor",
    "int_global_blue_traits_sensor_range",
    "int_global_blue_traits_waypoint",
    "int_global_blue_traits_active_camo",
    "int_global_blue_traits_visual_aura",
    "int_global_blue_traits_forced_color",
    "int_global_yellow_traits_damage_resistance",
    "int_global_yellow_traits_shield_multiplier",
    "int_global_yellow_traits_shield_recharge",
    "int_global_yellow_traits_headshot_immunity",
    "int_global_yellow_traits_vampirism",
    "int_global_yellow_traits_damage_modifier",
    "sid_global_yellow_traits_primary_weapon",
    "sid_global_yellow_traits_secondary_weapon",
    "int_global_yellow_traits_grenade_count",
    "int_global_yellow_traits_grenade_recharge",
    "int_global_yellow_traits_infinite_ammo",
    "int_global_yellow_traits_weapon_pickup",
    "int_global_yellow_traits_speed",
    "int_global_yellow_traits_gravity",
    "int_global_yellow_traits_vehicle_use",
    "int_global_yellow_traits_sensor",
    "int_global_yellow_traits_sensor_range",
    "int_global_yellow_traits_waypoint",
    "int_global_yellow_traits_active_camo",
    "int_global_yellow_traits_visual_aura",
    "int_global_yellow_traits_forced_color",
    "int_global_red_traits_duration",
    "int_global_blue_traits_duration",
    "int_global_yellow_traits_duration",
    "int_slayer_team_scoring",
    "int_slayer_score_to_win",
    "int_slayer_kill_points",
    "int_slayer_assist_points",
    "int_slayer_death_points",
    "int_slayer_suicide_points",
    "int_slayer_betrayal_points",
    "int_slayer_leader_points",
    "int_slayer_elimination_points",
    "int_slayer_assassination_points",
    "int_slayer_headshot_points",
    "int_slayer_melee_points",
    "int_slayer_sticky_points",
    "int_slayer_splatter_points",
    "int_slayer_spree_points",
    "int_slayer_leader_traits_damage_resistance",
    "int_slayer_leader_traits_shield_multiplier",
    "int_slayer_leader_traits_shield_recharge",
    "int_slayer_leader_traits_headshot_immunity",
    "int_slayer_leader_traits_vampirism",
    "int_slayer_leader_traits_damage_modifier",
    "sid_slayer_leader_traits_primary_weapon",
    "sid_slayer_leader_traits_secondary_weapon",
    "int_slayer_leader_traits_grenade_count",
    "int_slayer_leader_traits_grenade_recharge",
    "int_slayer_leader_traits_infinite_ammo",
    "int_slayer_leader_traits_weapon_pickup",
    "int_slayer_leader_traits_speed",
    "int_slayer_leader_traits_gravity",
    "int_slayer_leader_traits_vehicle_use",
    "int_slayer_leader_traits_sensor",
    "int_slayer_leader_traits_sensor_range",
    "int_slayer_leader_traits_waypoint",
    "int_slayer_leader_traits_active_camo",
    "int_slayer_leader_traits_visual_aura",
    "int_slayer_leader_traits_forced_color",
    "int_oddball_auto_pickup",
    "int_oddball_ball_effect",
    "int_oddball_team_scoring",
    "int_oddball_unknown",
    "int_oddball_score_to_win",
    "int_oddball_carrier_points",
    "int_oddball_kill_points",
    "int_oddball_ball_kill_points",
    "int_oddball_carrier_kill_points",
    "int_oddball_ball_count",
    "int_oddball_spawn_delay",
    "int_oddball_respawn_delay",
    "int_oddball_carrier_traits_damage_resistance",
    "int_oddball_carrier_traits_shield_multiplier",
    "int_oddball_carrier_traits_shield_recharge",
    "int_oddball_carrier_traits_headshot_immunity",
    "int_oddball_carrier_traits_vampirism",
    "int_oddball_carrier_traits_damage_modifier",
    "sid_oddball_carrier_traits_primary_weapon",
    "sid_oddball_carrier_traits_secondary_weapon",
    "int_oddball_carrier_traits_grenade_count",
    "int_oddball_carrier_traits_grenade_recharge",
    "int_oddball_carrier_traits_infinite_ammo",
    "int_oddball_carrier_traits_weapon_pickup",
    "int_oddball_carrier_traits_speed",
    "int_oddball_carrier_traits_gravity",
    "int_oddball_carrier_traits_vehicle_use",
    "int_oddball_carrier_traits_sensor",
    "int_oddball_carrier_traits_sensor_range",
    "int_oddball_carrier_traits_waypoint",
    "int_oddball_carrier_traits_active_camo",
    "int_oddball_carrier_traits_visual_aura",
    "int_oddball_carrier_traits_forced_color",
    "int_ctf_touch_return_time",
    "int_ctf_flag_at_home",
    "int_ctf_home_waypoint",
    "int_ctf_unknown_0",
    "int_ctf_unknown_1",
    "int_ctf_game_mode",
    "int_ctf_respawn_on_capture",
    "int_ctf_score_to_win",
    "int_ctf_sudden_death",
    "int_ctf_idle_return",
    "int_ctf_unknown_2",
    "int_ctf_unknown_3",
    "int_ctf_carrier_traits_damage_resistance",
    "int_ctf_carrier_traits_shield_multiplier",
    "int_ctf_carrier_traits_shield_recharge",
    "int_ctf_carrier_traits_headshot_immunity",
    "int_ctf_carrier_traits_vampirism",
    "int_ctf_carrier_traits_damage_modifier",
    "sid_ctf_carrier_traits_primary_weapon",
    "sid_ctf_carrier_traits_secondary_weapon",
    "int_ctf_carrier_traits_grenade_count",
    "int_ctf_carrier_traits_grenade_recharge",
    "int_ctf_carrier_traits_infinite_ammo",
    "int_ctf_carrier_traits_weapon_pickup",
    "int_ctf_carrier_traits_speed",
    "int_ctf_carrier_traits_gravity",
    "int_ctf_carrier_traits_vehicle_use",
    "int_ctf_carrier_traits_sensor",
    "int_ctf_carrier_traits_sensor_range",
    "int_ctf_carrier_traits_waypoint",
    "int_ctf_carrier_traits_active_camo",
    "int_ctf_carrier_traits_visual_aura",
    "int_ctf_carrier_traits_forced_color",
    "int_infection_respawn_on_haven_move",
    "int_infection_haven_movement",
    "int_infection_next_zombie",
    "int_infection_initial_count",
    "int_infection_haven_movement_time",
    "int_infection_kill_points",
    "int_infection_infect_points",
    "int_infection_haven_arrival_points",
    "int_infection_suicide_points",
    "int_infection_betrayal_points",
    "int_infection_last_man_standing_points",
    "int_infection_zombie_traits_damage_resistance",
    "int_infection_zombie_traits_shield_multiplier",
    "int_infection_zombie_traits_shield_recharge",
    "int_infection_zombie_traits_headshot_immunity",
    "int_infection_zombie_traits_vampirism",
    "int_infection_zombie_traits_damage_modifier",
    "sid_infection_zombie_traits_primary_weapon",
    "sid_infection_zombie_traits_secondary_weapon",
    "int_infection_zombie_traits_grenade_count",
    "int_infection_zombie_traits_grenade_recharge",
    "int_infection_zombie_traits_infinite_ammo",
    "int_infection_zombie_traits_weapon_pickup",
    "int_infection_zombie_traits_speed",
    "int_infection_zombie_traits_gravity",
    "int_infection_zombie_traits_vehicle_use",
    "int_infection_zombie_traits_sensor",
    "int_infection_zombie_traits_sensor_range",
    "int_infection_zombie_traits_waypoint",
    "int_infection_zombie_traits_active_camo",
    "int_infection_zombie_traits_visual_aura",
    "int_infection_zombie_traits_forced_color",
    "int_infection_alpha_zombie_traits_damage_resistance",
    "int_infection_alpha_zombie_traits_shield_multiplier",
    "int_infection_alpha_zombie_traits_shield_recharge",
    "int_infection_alpha_zombie_traits_headshot_immunity",
    "int_infection_alpha_zombie_traits_vampirism",
    "int_infection_alpha_zombie_traits_damage_modifier",
    "sid_infection_alpha_zombie_traits_primary_weapon",
    "sid_infection_alpha_zombie_traits_secondary_weapon",
    "int_infection_alpha_zombie_traits_grenade_count",
    "int_infection_alpha_zombie_traits_grenade_recharge",
    "int_infection_alpha_zombie_traits_infinite_ammo",
    "int_infection_alpha_zombie_traits_weapon_pickup",
    "int_infection_alpha_zombie_traits_speed",
    "int_infection_alpha_zombie_traits_gravity",
    "int_infection_alpha_zombie_traits_vehicle_use",
    "int_infection_alpha_zombie_traits_sensor",
    "int_infection_alpha_zombie_traits_sensor_range",
    "int_infection_alpha_zombie_traits_waypoint",
    "int_infection_alpha_zombie_traits_active_camo",
    "int_infection_alpha_zombie_traits_visual_aura",
    "int_infection_alpha_zombie_traits_forced_color",
    "int_infection_haven_traits_damage_resistance",
    "int_infection_haven_traits_shield_multiplier",
    "int_infection_haven_traits_shield_recharge",
    "int_infection_haven_traits_headshot_immunity",
    "int_infection_haven_traits_vampirism",
    "int_infection_haven_traits_damage_modifier",
    "sid_infection_haven_traits_primary_weapon",
    "sid_infection_haven_traits_secondary_weapon",
    "int_infection_haven_traits_grenade_count",
    "int_infection_haven_traits_grenade_recharge",
    "int_infection_haven_traits_infinite_ammo",
    "int_infection_haven_traits_weapon_pickup",
    "int_infection_haven_traits_speed",
    "int_infection_haven_traits_gravity",
    "int_infection_haven_traits_vehicle_use",
    "int_infection_haven_traits_sensor",
    "int_infection_haven_traits_sensor_range",
    "int_infection_haven_traits_waypoint",
    "int_infection_haven_traits_active_camo",
    "int_infection_haven_traits_visual_aura",
    "int_infection_haven_traits_forced_color",
    "int_infection_last_man_standing_traits_damage_resistance",
    "int_infection_last_man_standing_traits_shield_multiplier",
    "int_infection_last_man_standing_traits_shield_recharge",
    "int_infection_last_man_standing_traits_headshot_immunity",
    "int_infection_last_man_standing_traits_vampirism",
    "int_infection_last_man_standing_traits_damage_modifier",
    "sid_infection_last_man_standing_traits_primary_weapon",
    "sid_infection_last_man_standing_traits_secondary_weapon",
    "int_infection_last_man_standing_traits_grenade_count",
    "int_infection_last_man_standing_traits_grenade_recharge",
    "int_infection_last_man_standing_traits_infinite_ammo",
    "int_infection_last_man_standing_traits_weapon_pickup",
    "int_infection_last_man_standing_traits_speed",
    "int_infection_last_man_standing_traits_gravity",
    "int_infection_last_man_standing_traits_vehicle_use",
    "int_infection_last_man_standing_traits_sensor",
    "int_infection_last_man_standing_traits_sensor_range",
    "int_infection_last_man_standing_traits_waypoint",
    "int_infection_last_man_standing_traits_active_camo",
    "int_infection_last_man_standing_traits_visual_aura",
    "int_infection_last_man_standing_traits_forced_color",
    "int_koth_opaque_hill",
    "int_koth_score_to_win",
    "int_koth_team_scoring",
    "int_koth_moving_hill",
    "int_koth_moving_hill_order",
    "int_koth_inside_hill_points",
    "int_koth_outside_hill_points",
    "int_koth_uncontested_points",
    "int_koth_kill_points",
    "int_koth_hill_traits_damage_resistance",
    "int_koth_hill_traits_shield_multiplier",
    "int_koth_hill_traits_shield_recharge",
    "int_koth_hill_traits_headshot_immunity",
    "int_koth_hill_traits_vampirism",
    "int_koth_hill_traits_damage_modifier",
    "sid_koth_hill_traits_primary_weapon",
    "sid_koth_hill_traits_secondary_weapon",
    "int_koth_hill_traits_grenade_count",
    "int_koth_hill_traits_grenade_recharge",
    "int_koth_hill_traits_infinite_ammo",
    "int_koth_hill_traits_weapon_pickup",
    "int_koth_hill_traits_speed",
    "int_koth_hill_traits_gravity",
    "int_koth_hill_traits_vehicle_use",
    "int_koth_hill_traits_sensor",
    "int_koth_hill_traits_sensor_range",
    "int_koth_hill_traits_waypoint",
    "int_koth_hill_traits_active_camo",
    "int_koth_hill_traits_visual_aura",
    "int_koth_hill_traits_forced_color",
    "int_territories_capture_time",
    "int_territories_game_mode",
    "int_territories_lock_on_capture",
    "int_territories_sudden_death",
    "int_territories_spawn_on_capture",
    "int_territories_defender_traits_damage_resistance",
    "int_territories_defender_traits_shield_multiplier",
    "int_territories_defender_traits_shield_recharge",
    "int_territories_defender_traits_headshot_immunity",
    "int_territories_defender_traits_vampirism",
    "int_territories_defender_traits_damage_modifier",
    "sid_territories_defender_traits_primary_weapon",
    "sid_territories_defender_traits_secondary_weapon",
    "int_territories_defender_traits_grenade_count",
    "int_territories_defender_traits_grenade_recharge",
    "int_territories_defender_traits_infinite_ammo",
    "int_territories_defender_traits_weapon_pickup",
    "int_territories_defender_traits_speed",
    "int_territories_defender_traits_gravity",
    "int_territories_defender_traits_vehicle_use",
    "int_territories_defender_traits_sensor",
    "int_territories_defender_traits_sensor_range",
    "int_territories_defender_traits_waypoint",
    "int_territories_defender_traits_active_camo",
    "int_territories_defender_traits_visual_aura",
    "int_territories_defender_traits_forced_color",
    "int_territories_attacker_traits_damage_resistance",
    "int_territories_attacker_traits_shield_multiplier",
    "int_territories_attacker_traits_shield_recharge",
    "int_territories_attacker_traits_headshot_immunity",
    "int_territories_attacker_traits_vampirism",
    "int_territories_attacker_traits_damage_modifier",
    "sid_territories_attacker_traits_primary_weapon",
    "sid_territories_attacker_traits_secondary_weapon",
    "int_territories_attacker_traits_grenade_count",
    "int_territories_attacker_traits_grenade_recharge",
    "int_territories_attacker_traits_infinite_ammo",
    "int_territories_attacker_traits_weapon_pickup",
    "int_territories_attacker_traits_speed",
    "int_territories_attacker_traits_gravity",
    "int_territories_attacker_traits_vehicle_use",
    "int_territories_attacker_traits_sensor",
    "int_territories_attacker_traits_sensor_range",
    "int_territories_attacker_traits_waypoint",
    "int_territories_attacker_traits_active_camo",
    "int_territories_attacker_traits_visual_aura",
    "int_territories_attacker_traits_forced_color",
    "int_vip_score_to_win",
    "int_vip_game_mode",
    "int_vip_zone_count",
    "int_vip_end_on_death",
    "int_vip_kill_points",
    "int_vip_takedown_points",
    "int_vip_vip_kill_points",
    "int_vip_death_points",
    "int_vip_zone_arrival_points",
    "int_vip_suicide_points",
    "int_vip_vip_betrayal_points",
    "int_vip_betrayal_points",
    "int_vip_selection",
    "int_vip_zone_movement",
    "int_vip_zone_order",
    "int_vip_influence_radius",
    "int_vip_team_traits_damage_resistance",
    "int_vip_team_traits_shield_multiplier",
    "int_vip_team_traits_shield_recharge",
    "int_vip_team_traits_headshot_immunity",
    "int_vip_team_traits_vampirism",
    "int_vip_team_traits_damage_modifier",
    "sid_vip_team_traits_primary_weapon",
    "sid_vip_team_traits_secondary_weapon",
    "int_vip_team_traits_grenade_count",
    "int_vip_team_traits_grenade_recharge",
    "int_vip_team_traits_infinite_ammo",
    "int_vip_team_traits_weapon_pickup",
    "int_vip_team_traits_speed",
    "int_vip_team_traits_gravity",
    "int_vip_team_traits_vehicle_use",
    "int_vip_team_traits_sensor",
    "int_vip_team_traits_sensor_range",
    "int_vip_team_traits_waypoint",
    "int_vip_team_traits_active_camo",
    "int_vip_team_traits_visual_aura",
    "int_vip_team_traits_forced_color",
    "int_vip_influence_traits_damage_resistance",
    "int_vip_influence_traits_shield_multiplier",
    "int_vip_influence_traits_shield_recharge",
    "int_vip_influence_traits_headshot_immunity",
    "int_vip_influence_traits_vampirism",
    "int_vip_influence_traits_damage_modifier",
    "sid_vip_influence_traits_primary_weapon",
    "sid_vip_influence_traits_secondary_weapon",
    "int_vip_influence_traits_grenade_count",
    "int_vip_influence_traits_grenade_recharge",
    "int_vip_influence_traits_infinite_ammo",
    "int_vip_influence_traits_weapon_pickup",
    "int_vip_influence_traits_speed",
    "int_vip_influence_traits_gravity",
    "int_vip_influence_traits_vehicle_use",
    "int_vip_influence_traits_sensor",
    "int_vip_influence_traits_sensor_range",
    "int_vip_influence_traits_waypoint",
    "int_vip_influence_traits_active_camo",
    "int_vip_influence_traits_visual_aura",
    "int_vip_influence_traits_forced_color",
    "int_vip_vip_traits_damage_resistance",
    "int_vip_vip_traits_shield_multiplier",
    "int_vip_vip_traits_shield_recharge",
    "int_vip_vip_traits_headshot_immunity",
    "int_vip_vip_traits_vampirism",
    "int_vip_vip_traits_damage_modifier",
    "sid_vip_vip_traits_primary_weapon",
    "sid_vip_vip_traits_secondary_weapon",
    "int_vip_vip_traits_grenade_count",
    "int_vip_vip_traits_grenade_recharge",
    "int_vip_vip_traits_infinite_ammo",
    "int_vip_vip_traits_weapon_pickup",
    "int_vip_vip_traits_speed",
    "int_vip_vip_traits_gravity",
    "int_vip_vip_traits_vehicle_use",
    "int_vip_vip_traits_sensor",
    "int_vip_vip_traits_sensor_range",
    "int_vip_vip_traits_waypoint",
    "int_vip_vip_traits_active_camo",
    "int_vip_vip_traits_visual_aura",
    "int_vip_vip_traits_forced_color",
    "int_juggernaut_score_to_win",
    "int_juggernaut_kill_points",
    "int_juggernaut_juggy_kill_points",
    "int_juggernaut_kill_as_juggy_points",
    "int_juggernaut_zone_arrival_points",
    "int_juggernaut_suicide_points",
    "int_juggernaut_betrayal_points",
    "int_juggernaut_initial_juggy_selection",
    "int_juggernaut_next_juggy_selection",
    "int_juggernaut_allied_against_juggy",
    "int_juggernaut_next_juggy_delay",
    "int_juggernaut_zones_enabled",
    "int_juggernaut_zone_movement",
    "int_juggernaut_zone_order",
    "int_juggernaut_respawn_on_lone_juggy",
    "int_juggernaut_juggernaut_traits_damage_resistance",
    "int_juggernaut_juggernaut_traits_shield_multiplier",
    "int_juggernaut_juggernaut_traits_shield_recharge",
    "int_juggernaut_juggernaut_traits_headshot_immunity",
    "int_juggernaut_juggernaut_traits_vampirism",
    "int_juggernaut_juggernaut_traits_damage_modifier",
    "sid_juggernaut_juggernaut_traits_primary_weapon",
    "sid_juggernaut_juggernaut_traits_secondary_weapon",
    "int_juggernaut_juggernaut_traits_grenade_count",
    "int_juggernaut_juggernaut_traits_grenade_recharge",
    "int_juggernaut_juggernaut_traits_infinite_ammo",
    "int_juggernaut_juggernaut_traits_weapon_pickup",
    "int_juggernaut_juggernaut_traits_speed",
    "int_juggernaut_juggernaut_traits_gravity",
    "int_juggernaut_juggernaut_traits_vehicle_use",
    "int_juggernaut_juggernaut_traits_sensor",
    "int_juggernaut_juggernaut_traits_sensor_range",
    "int_juggernaut_juggernaut_traits_waypoint",
    "int_juggernaut_juggernaut_traits_active_camo",
    "int_juggernaut_juggernaut_traits_visual_aura",
    "int_juggernaut_juggernaut_traits_forced_color",
    "int_unknown_traits_damage_resistance",
    "int_unknown_traits_shield_multiplier",
    "int_unknown_traits_shield_recharge",
    "int_unknown_traits_headshot_immunity",
    "int_unknown_traits_vampirism",
    "int_unknown_traits_damage_modifier",
    "sid_unknown_traits_primary_weapon",
    "sid_unknown_traits_secondary_weapon",
    "int_unknown_traits_grenade_count",
    "int_unknown_traits_grenade_recharge",
    "int_unknown_traits_infinite_ammo",
    "int_unknown_traits_weapon_pickup",
    "int_unknown_traits_speed",
    "int_unknown_traits_gravity",
    "int_unknown_traits_vehicle_use",
    "int_unknown_traits_sensor",
    "int_unknown_traits_sensor_range",
    "int_unknown_traits_waypoint",
    "int_unknown_traits_active_camo",
    "int_unknown_traits_visual_aura",
    "int_unknown_traits_forced_color",
    "int_assault_arm_time",
    "int_assault_disarm_time",
    "int_assault_fuse_time",
    "int_assault_reset_time",
    "int_assault_carrier_traits_damage_resistance",
    "int_assault_carrier_traits_shield_multiplier",
    "int_assault_carrier_traits_shield_recharge",
    "int_assault_carrier_traits_headshot_immunity",
    "int_assault_carrier_traits_vampirism",
    "int_assault_carrier_traits_damage_modifier",
    "sid_assault_carrier_traits_primary_weapon",
    "sid_assault_carrier_traits_secondary_weapon",
    "int_assault_carrier_traits_grenade_count",
    "int_assault_carrier_traits_grenade_recharge",
    "int_assault_carrier_traits_infinite_ammo",
    "int_assault_carrier_traits_weapon_pickup",
    "int_assault_carrier_traits_speed",
    "int_assault_carrier_traits_gravity",
    "int_assault_carrier_traits_vehicle_use",
    "int_assault_carrier_traits_sensor",
    "int_assault_carrier_traits_sensor_range",
    "int_assault_carrier_traits_waypoint",
    "int_assault_carrier_traits_active_camo",
    "int_assault_carrier_traits_visual_aura",
    "int_assault_carrier_traits_forced_color",
    "int_assault_unknown",
    "int_assault_game_mode",
    "int_assault_reset_on_disarm",
    "int_assault_enemy_bomb_waypoint",
    "int_assault_score_to_win",
    "int_assault_sudden_death",
    "int_editor_open_channel_voice",
    "int_editor_edit_mode",
    "int_editor_respawn_time",
    "int_editor_editor_traits_damage_resistance",
    "int_editor_editor_traits_shield_multiplier",
    "int_editor_editor_traits_shield_recharge",
    "int_editor_editor_traits_headshot_immunity",
    "int_editor_editor_traits_vampirism",
    "int_editor_editor_traits_damage_modifier",
    "sid_editor_editor_traits_primary_weapon",
    "sid_editor_editor_traits_secondary_weapon",
    "int_editor_editor_traits_grenade_count",
    "int_editor_editor_traits_grenade_recharge",
    "int_editor_editor_traits_infinite_ammo",
    "int_editor_editor_traits_weapon_pickup",
    "int_editor_editor_traits_speed",
    "int_editor_editor_traits_gravity",
    "int_editor_editor_traits_vehicle_use",
    "int_editor_editor_traits_sensor",
    "int_editor_editor_traits_sensor_range",
    "int_editor_editor_traits_waypoint",
    "int_editor_editor_traits_active_camo",
    "int_editor_editor_traits_visual_aura",
    "int_editor_editor_traits_forced_color",
    "int_global_traits_template_damage_resistance",
    "int_global_traits_template_shield_multiplier",
    "int_global_traits_template_shield_recharge",
    "int_global_traits_template_headshot_immunity",
    "int_global_traits_template_vampirism",
    "int_global_traits_template_damage_modifier",
    "sid_global_traits_template_primary_weapon",
    "sid_global_traits_template_secondary_weapon",
    "int_global_traits_template_grenade_count",
    "int_global_traits_template_grenade_recharge",
    "int_global_traits_template_infinite_ammo",
    "int_global_traits_template_weapon_pickup",
    "int_global_traits_template_speed",
    "int_global_traits_template_gravity",
    "int_global_traits_template_vehicle_use",
    "int_global_traits_template_sensor",
    "int_global_traits_template_sensor_range",
    "int_global_traits_template_waypoint",
    "int_global_traits_template_active_camo",
    "int_global_traits_template_visual_aura",
    "int_global_traits_template_forced_color",
    )


sily_text_value_pair = Struct("text_value_pair",
    Bool8("flags",
        "default_setting",
        "unchanged_setting",
        ),
    SEnum8("expected_value_type", *sily_text_value_pair_expected_value_type),
    SInt16("unknown", VISIBLE=False),
    SInt32("integer_value"),
    h3_string_id("stringid_value"),
    h3_string_id("name"),
    h3_string_id("description"),
    ENDIAN=">", SIZE=20
    )


sily_body = Struct("tagdata",
    SEnum32("parameter", *sily_parameter),
    h3_string_id("name"),
    h3_string_id("description"),
    h3_reflexive("text_value_pairs", sily_text_value_pair),
    ENDIAN=">", SIZE=24
    )


def get():
    return sily_def

sily_def = TagDef("sily",
    h3_blam_header('sily'),
    sily_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["sily"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...os_v3_hek.defs.vehi import *

#import and use the open saucified obje and unit attrs
from .obje import *
from .unit import *

# replace the object_type enum one that uses
# the correct default value for this object
obje_attrs = dict(obje_attrs)
obje_attrs[0] = dict(obje_attrs[0], DEFAULT=1)

vehi_body = dict(vehi_body)
vehi_body[0] = obje_attrs
vehi_body[1] = unit_attrs

def get():
    return vehi_def

vehi_def = TagDef("vehi",
    blam_header('vehi'),
    vehi_body,

    ext=".vehicle", endian=">", tag_cls=ObjeTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ..common_descs import *
from supyr_struct.defs.tag_def import TagDef

def get():
    return t3ed_def

t3ed_def = TagDef("t3ed",
    blam_header('t3ed'),
    Struct('tagdata',
        ),

    ext=".tech_3_effect_data", endian=">"
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from reclaimer.constants import *
from reclaimer.util import fourcc_to_int

# maps tag class four character codes(fccs) in
# their string encoding to their int encoding.
stubbs_tag_class_fcc_to_be_int = {}
stubbs_tag_class_fcc_to_le_int = {}
# maps tag class four character codes(fccs) in
# their int encoding to their string encoding.
stubbs_tag_class_be_int_to_fcc = {}
stubbs_tag_class_le_int_to_fcc = {}


stubbs_tag_class_fcc_to_ext = dict(
    terr="terrain",
    vege="vegetation",
    imef="image_effect",
    **tag_class_fcc_to_ext
    )

for tag_cls in stubbs_tag_class_fcc_to_ext:
    stubbs_tag_class_fcc_to_be_int[tag_cls] = fourcc_to_int(tag_cls)
    stubbs_tag_class_be_int_to_fcc[fourcc_to_int(tag_cls)] = tag_cls
    stubbs_tag_class_fcc_to_le_int[tag_cls] = fourcc_to_int(tag_cls)
    stubbs_tag_class_le_int_to_fcc[fourcc_to_int(tag_cls)] = tag_cls



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.actv import *

# TODO use some supyr trickery to make this nicer.

# grenades descriptor index in actv_body is 10
# grenade_type descriptor index in grenades_desc is 1

# replace the grenade_type descriptor with one
# that uses open sauce's extra grenade slots
actv_body = dict(actv_body)
grenades_desc = actv_body[10] = dict(actv_body[10])
grenades_desc[1] = SEnum16("grenade_type", *grenade_types_os)

def get():
    return actv_def

actv_def = TagDef("actv",
    blam_header('actv'),
    actv_body,

    ext=".actor_variant", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


shit_body = Struct("tagdata",
    h3_dependency("unknown_0"),
    h3_dependency("unknown_1"),
    Array("unknown_array", SUB_STRUCT=Float("unknown"), SIZE=9, VISIBLE=False),
    color_rgb_float("color_0"),
    Float("unknown_2", VISIBLE=False),
    color_rgb_float("color_1"),
    Float("unknown_3", VISIBLE=False),
    color_rgb_float("color_2"),
    Float("unknown_4", VISIBLE=False),
    color_rgb_float("color_3"),
    Float("unknown_5", VISIBLE=False),
    color_rgb_float("color_4"),
    Float("unknown_6", VISIBLE=False),
    color_rgb_float("color_5"),
    Float("unknown_7", VISIBLE=False),
    ENDIAN=">", SIZE=164
    )


def get():
    return shit_def

shit_def = TagDef("shit",
    h3_blam_header('shit'),
    shit_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["shit"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Lord Zedd
# 	Standardizing
# revision: 3		author: Lehvak
# 	Mapped Bit 7
# revision: 4		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


ligh_body = Struct("tagdata",
    Bool32("flags",
        ("no_shadow", 1 << 1),
        "only_render_in_first_person",
        "only_render_in_third_person",
        ("snap_to_first_person_camera_req_only_first_person", 1 << 7),
        ),
    SEnum16("type", *scnr_light_volume_type_1),
    SInt16("unknown_0", VISIBLE=False),
    Float("light_range"),
    Float("near_width"),
    Float("height_stretch"),
    float_rad("field_of_view"),
    h3_string_id("function_name_0"),
    h3_string_id("function_name_1"),
    SInt16("unknown_1", VISIBLE=False),
    SInt16("unknown_2", VISIBLE=False),
    BytesRaw("unknown_3", SIZE=4, VISIBLE=False),
    h3_rawdata_ref("function_0"),
    h3_string_id("function_name_2"),
    h3_string_id("function_name_3"),
    SInt16("unknown_4", VISIBLE=False),
    SInt16("unknown_5", VISIBLE=False),
    BytesRaw("unknown_6", SIZE=4, VISIBLE=False),
    h3_rawdata_ref("function_1"),
    h3_dependency("gel_map"),
    Array("unknown_array_0", SUB_STRUCT=Float("unknown"), SIZE=4, VISIBLE=False),
    Array("unknown_array_1", SUB_STRUCT=SInt8("unknown"), SIZE=4, VISIBLE=False),
    h3_dependency("lens_flare"),
    ENDIAN=">", SIZE=148
    )


def get():
    return ligh_def

ligh_def = TagDef("ligh",
    h3_blam_header('ligh'),
    ligh_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["ligh"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from struct import unpack

from supyr_struct.defs.constants import *
from supyr_struct.util import fourcc_to_int
from binilla.constants import *

# some reflexives are so massive that it's significantly faster to treat them
# as raw data and just byteswap them using precalculated offsets and sizes
RAW_REFLEXIVE_INFO = "RAW_REFLEXIVE_INFO"
STRINGID_IDX_BITS = "STRINGID_IDX_BITS"
STRINGID_SET_BITS = "STRINGID_SET_BITS"
STRINGID_LEN_BITS = "STRINGID_LEN_BITS"
COLOR_CHANNELS = "COLOR_CHANNELS"
IGNORE_SAFE_MODE = "IGNORE_SAFE_MODE"

def inject_halo_constants():
    # add the new descriptor keywords to the sets
    add_desc_keywords(RAW_REFLEXIVE_INFO, COLOR_CHANNELS,
                      MAX, MIN, IGNORE_SAFE_MODE,
                      STRINGID_IDX_BITS, STRINGID_SET_BITS, STRINGID_LEN_BITS)


PCDEMO_INDEX_MAGIC          = 0x4BF10000
PC_INDEX_MAGIC              = 0x40440000
CE_INDEX_MAGIC              = 0x40440000
ANNIVERSARY_INDEX_MAGIC     = 0x004B8000
XBOX_INDEX_MAGIC            = 0x803A6000
STUBBS_INDEX_MAGIC          = 0x8038B000
SHADOWRUN_PROTO_INDEX_MAGIC = 0x8069E000
H2_XBOX_INDEX_MAGIC         = 0x80061000


map_build_dates = {
    "stubbs":          "400",
    "stubbspc":        "",
    "shadowrun_proto": "01.12.07.0132",
    "halo1xboxdemo":   "",
    "halo1xbox":       "01.10.12.2276",
    "halo1pcdemo":     "01.00.00.0576",
    "halo1anni":       "01.00.01.0563",
    "halo1ce":         "01.00.00.0609",
    "halo1yelo":       "01.00.00.0609",
    "halo1vap":        "01.00.00.0609",
    "halo1pc":         "01.00.00.0564",
    "halo2alpha":      "02.01.07.4998",
    "halo2beta":       "02.06.28.07902",
    "halo2epsilon":    "02.08.28.09214",
    "halo2xbox":       "02.09.27.09809",
    "halo2vista":      "11081.07.04.30.0934.main",
    "halo3beta":       "",  # SET THIS
    "halo3":           "11855.07.08.20.2317.halo3_ship",
    "halo3odst":       "13895.09.04.27.2201.atlas_relea",
    "haloreachbeta":   "11860.10.07.24.0147.omaha_relea",
    "haloreach":       "",  # SET THIS
    "halo4":           "20810.12.09.22.1647.main",
    "halo4nettest":    "",  # SET THIS
    "halo5":           "",  # SET THIS
    }

map_versions = {
    "stubbs":          5,
    "stubbspc":        5,
    "shadowrun_proto": 5,
    "halo1xboxdemo":   5,
    "halo1xbox":       5,
    "halo1pcdemo":     6,
    "halo1pc":         7,
    "halo1anni":       7,
    "halo1ce":         609,
    "halo1yelo":       609,
    "halo1vap":        134,
    "halo2alpha":      7,
    "halo2beta":       8,
    "halo2epsilon":    8,
    "halo2xbox":       8,
    "halo2vista":      8,
    "halo3beta":       9,
    "halo3":           11,
    "halo3odst":       11,
    "haloreachbeta":   12,
    "haloreach":       12,
    "halo4":           12,
    #"halo4nettest":    ????,
    #"halo5":           ????,
    }

GEN_1_HALO_ENGINES = ("halo1xboxdemo", "halo1xbox",
                      "halo1ce", "halo1vap", "halo1yelo",
                      "halo1pcdemo", "halo1pc", "halo1anni", )

GEN_1_ENGINES = GEN_1_HALO_ENGINES + (
    "stubbs", "stubbspc", "shadowrun_proto", )

GEN_2_ENGINES = ("halo2alpha", "halo2beta", "halo2epsilon",
                 "halo2xbox", "halo2vista", )

GEN_3_ENGINES = ("halo3", "halo3odst", "halo3beta",
                 "haloreachbeta", "haloreach",
                 "halo4", "halo4nettest", "halo5", )

# magic is actually the virtual address the map is loaded at. Halo 3 and
# beyond instead partition the map into sections with a virtual address for
# each section, meaning there is a "magic" for different parts of each map.
map_magics = {
    "stubbs":          STUBBS_INDEX_MAGIC,
    "stubbspc":        PC_INDEX_MAGIC,
    "shadowrun_proto": SHADOWRUN_PROTO_INDEX_MAGIC,
    "halo1xboxdemo":   XBOX_INDEX_MAGIC,
    "halo1xbox":       XBOX_INDEX_MAGIC,
    "halo1pcdemo":     PCDEMO_INDEX_MAGIC,
    "halo1pc":         PC_INDEX_MAGIC,
    "halo1anni":       ANNIVERSARY_INDEX_MAGIC,
    "halo1ce":         CE_INDEX_MAGIC,
    "halo1yelo":       CE_INDEX_MAGIC,
    "halo1vap":        CE_INDEX_MAGIC,
    "halo2alpha":      H2_XBOX_INDEX_MAGIC,
    "halo2beta":       H2_XBOX_INDEX_MAGIC,
    "halo2epsilon":    H2_XBOX_INDEX_MAGIC,
    "halo2xbox":       H2_XBOX_INDEX_MAGIC,
    }

# bitmap types
TYPE_2D = 0
TYPE_3D = 1
TYPE_CUBEMAP = 2
TYPE_WHITE = 3

# bitmap formats
FORMAT_A8 = 0
FORMAT_Y8 = 1
FORMAT_AY8 = 2
FORMAT_A8Y8 = 3
FORMAT_R5G6B5 = 6
FORMAT_A1R5G5B5 = 8
FORMAT_A4R4G4B4 = 9
FORMAT_X8R8G8B8 = 10
FORMAT_A8R8G8B8 = 11
FORMAT_DXT1 = 14
FORMAT_DXT3 = 15
FORMAT_DXT5 = 16
FORMAT_P8_BUMP = 17
FORMAT_P8 = 18
FORMAT_A32R32G32B32F = 19
FORMAT_R32G32B32F = 20
FORMAT_R16G16B16F = 21
FORMAT_V8U8 = 22
FORMAT_G8B8 = 23
FORMAT_DXN = 33
FORMAT_CTX1 = 34
FORMAT_DXT3A = 35
FORMAT_DXT3Y = 36
FORMAT_DXT5A = 37
FORMAT_DXT5Y = 38
FORMAT_DXT5AY = 39

DXT_FORMATS = (FORMAT_DXT1, FORMAT_DXT3, FORMAT_DXT5)

PALLETIZED_FORMATS = (FORMAT_P8_BUMP, )

# These name maps must match the constants found in arbytmap
TYPE_NAME_MAP = ("2D", "3D", "CUBE", "WHITE")

# this map corrosponds to the bitmap formats
# found in the "format" enum in the bitmap tag
# NOTE: These names must be unique, as they are used to map arbytmap
# formats to halo bitmap format enum values. P8 needs to be its own
# format which is just a variant of A8R8G8B8 so we know its P8.
FORMAT_NAME_MAP = (
    "A8", "L8", "AL8", "A8L8",
    "UNUSED1", "UNUSED2",
    "R5G6B5",  "UNUSED3", "A1R5G5B5", "A4R4G4B4",
    "X8R8G8B8", "A8R8G8B8",
    "UNUSED4", "UNUSED5",
    "DXT1", "DXT3", "DXT5", "P8-BUMP", "P8",
    "A32R32G32B32F", "R32G32B32F", "R16G16B16F",
    "V8U8", "G8B8", "UNUSED6", "UNUSED7", "UNUSED8",
    "UNUSED9", "UNUSED10", "UNUSED11", "UNUSED12", "UNUSED13",
    "UNUSED14", "DXN", "CTX1", "DXT3A", "DXT3Y", "DXT5A", "DXT5Y", "DXT5AY")

I_FORMAT_NAME_MAP = {}
for i in range(len(FORMAT_NAME_MAP)):
    if i not in I_FORMAT_NAME_MAP:
        I_FORMAT_NAME_MAP[FORMAT_NAME_MAP[i]] = i

#each bitmap's number of bytes must be a multiple of 512
BITMAP_PADDING = 512
#each sub-bitmap(cubemap face) must be a multiple of 128 bytes
CUBEMAP_PADDING = 128

#I cant imagine Halo allowing any one field even close to this many
#indices, though I have seen some open sauce stuff go over 180,000.
MAX_REFLEXIVE_COUNT = 2**31-1

MAX_TAG_PATH_LEN = 254

# maps tag class four character codes(fccs) in
# their string encoding to their int encoding.
tag_class_fcc_to_be_int = {}
tag_class_fcc_to_le_int = {}
# maps tag class four character codes(fccs) in
# their int encoding to their string encoding.
tag_class_be_int_to_fcc = {}
tag_class_le_int_to_fcc = {}

# maps tag class four character codes to the tags file extension
tag_class_fcc_to_ext = {
    'actr': "actor",
    'actv': "actor_variant",
    'ant!': "antenna",
    'bipd': "biped",
    'bitm': "bitmap",
    'trak': "camera_track",
    'colo': "color_table",
    'cdmg': "continuous_damage_effect",
    'cont': "contrail",
    'jpt!': "damage_effect",
    'deca': "decal",
    'udlg': "dialogue",
    'dobc': "detail_object_collection",
    'devi': "device",
    'ctrl': "device_control",
    'lifi': "device_light_fixture",
    'mach': "device_machine",
    'effe': "effect",
    'eqip': "equipment",
    'flag': "flag",
    'fog ': "fog",
    'font': "font",
    'garb': "garbage",
    'mod2': "gbxmodel",
    'matg': "globals",
    'glw!': "glow",
    'grhi': "grenade_hud_interface",
    'hudg': "hud_globals",
    'hmt ': "hud_message_text",
    'hud#': "hud_number",
    'devc': "input_device_defaults",
    'item': "item",
    'itmc': "item_collection",
    'lens': "lens_flare",
    'ligh': "light",
    'mgs2': "light_volume",
    'elec': "lightning",
    'foot': "material_effects",
    'metr': "meter",
    'mode': "model",
    'antr': "model_animations",
    'coll': "model_collision_geometry",
    'mply': "multiplayer_scenario_description",
    'obje': "object",
    'part': "particle",
    'pctl': "particle_system",
    'phys': "physics",
    'plac': "placeholder",
    'pphy': "point_physics",
    'ngpr': "preferences_network_game",
    'proj': "projectile",
    'scnr': "scenario",
    'sbsp': "scenario_structure_bsp",
    'scen': "scenery",
    'snd!': "sound",
    'snde': "sound_environment",
    'lsnd': "sound_looping",
    'ssce': "sound_scenery",
    'boom': "spheroid",
    'shdr': "shader",
    'schi': "shader_transparent_chicago",
    'scex': "shader_transparent_chicago_extended",
    'sotr': "shader_transparent_generic",
    'senv': "shader_environment",
    'sgla': "shader_transparent_glass",
    'smet': "shader_transparent_meter",
    'soso': "shader_model",
    'spla': "shader_transparent_plasma",
    'swat': "shader_transparent_water",
    'sky ': "sky",
    'str#': "string_list",
    'tagc': "tag_collection",
    'Soul': "ui_widget_collection",
    'DeLa': "ui_widget_definition",
    'ustr': "unicode_string_list",
    'unit': "unit",
    'unhi': "unit_hud_interface",
    'vehi': "vehicle",
    'vcky': "virtual_keyboard",
    'weap': "weapon",
    'wphi': "weapon_hud_interface",
    'rain': "weather_particle_system",
    'wind': "wind",
    }
tag_class_ext_to_fcc = {}

for tag_cls in tag_class_fcc_to_ext:
    tag_class_ext_to_fcc[tag_class_fcc_to_ext[tag_cls]] = tag_cls
    tag_class_fcc_to_be_int[tag_cls] = fourcc_to_int(tag_cls, 'big')
    tag_class_be_int_to_fcc[fourcc_to_int(tag_cls, 'big')] = tag_cls
    tag_class_fcc_to_le_int[tag_cls] = fourcc_to_int(tag_cls)
    tag_class_le_int_to_fcc[fourcc_to_int(tag_cls)] = tag_cls


################################
# Open Sauce related constants #
################################
tag_class_fcc_to_ext_os = {
    'efpp': "effect_postprocess",
    'efpc': "effect_postprocess_collection",
    'efpg': "effect_postprocess_generic",
    'eqhi': "equipment_hud_interface",
    'magy': "model_animations_yelo",
    'unic': "multilingual_unicode_string_list",
    'yelo': "project_yellow",
    'gelo': "project_yellow_globals",
    'gelc': "project_yellow_globals_cv",  # removed in OS v4
    'shpp': "shader_postprocess",
    'shpg': "shader_postprocess_generic",
    'sppg': "shader_postprocess_globals",
    'sidy': "string_id_yelo",
    'tag+': "tag_database",
    'sily': "text_value_pair_definition",

    # added in OS v4
    'avti': "actor_variant_transform_in",
    'avto': "actor_variant_transform_out",
    'avtc': "actor_variant_transform_collection",
    }

tag_class_fcc_to_ext_os.update(tag_class_fcc_to_ext)
tag_class_ext_to_fcc_os = {}

for tag_cls in tag_class_fcc_to_ext_os:
    tag_class_ext_to_fcc_os[tag_class_fcc_to_ext_os[tag_cls]] = tag_cls

# maps open sauce tag class four character codes(fccs)
# in their string encoding to their int encoding.
tag_class_fcc_to_be_int_os = {}
tag_class_fcc_to_le_int_os = {}
# maps open sauce tag class four character codes(fccs)
# in their int encoding to their string encoding.
tag_class_be_int_to_fcc_os = {}
tag_class_le_int_to_fcc_os = {}

for tag_cls in tag_class_fcc_to_ext_os:
    tag_class_fcc_to_be_int_os[tag_cls] = fourcc_to_int(tag_cls, 'big')
    tag_class_be_int_to_fcc_os[fourcc_to_int(tag_cls, 'big')] = tag_cls
    tag_class_fcc_to_le_int_os[tag_cls] = fourcc_to_int(tag_cls)
    tag_class_le_int_to_fcc_os[fourcc_to_int(tag_cls)] = tag_cls



#######################################
# Stubbs the Zombie related constants #
#######################################
tag_class_fcc_to_ext_stubbs = {
    'imef': "image_effect",
    'vege': "vegetation",
    'terr': "terrain",  # as kornman said, i dont fucking know
    }

tag_class_fcc_to_ext_stubbs.update(tag_class_fcc_to_ext)
tag_class_ext_to_fcc_stubbs = {}

for tag_cls in tag_class_fcc_to_ext_stubbs:
    tag_class_ext_to_fcc_stubbs[tag_class_fcc_to_ext_stubbs[tag_cls]] = tag_cls

# maps open sauce tag class four character codes(fccs)
# in their string encoding to their int encoding.
tag_class_fcc_to_be_int_stubbs = {}
tag_class_fcc_to_le_int_stubbs = {}
# maps open sauce tag class four character codes(fccs)
# in their int encoding to their string encoding.
tag_class_be_int_to_fcc_stubbs = {}
tag_class_le_int_to_fcc_stubbs = {}

for tag_cls in tag_class_fcc_to_ext_stubbs:
    tag_class_fcc_to_be_int_stubbs[tag_cls] = fourcc_to_int(tag_cls, 'big')
    tag_class_be_int_to_fcc_stubbs[fourcc_to_int(tag_cls, 'big')] = tag_cls
    tag_class_fcc_to_le_int_stubbs[tag_cls] = fourcc_to_int(tag_cls)
    tag_class_le_int_to_fcc_stubbs[fourcc_to_int(tag_cls)] = tag_cls



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

class MapPointerConverter:
    '''Handles converting virtual pointers to/from file pointers in maps.'''
    class PageInfo:
        def __init__(self, v_addr=0, f_addr=0, size=0):
            self.v_addr = v_addr
            self.f_addr = f_addr
            self.size = size
            self.v_addr_range = range(self.v_addr, self.v_addr + self.size)
            self.f_addr_range = range(self.f_addr, self.f_addr + self.size)

    _page_infos = ()

    def __init__(self, *pages):
        self._page_infos = []
        for v_addr, f_addr, size in pages:
            self.add_page_info(v_addr, f_addr, size)

    @property
    def mapped_size(self):
        size = 0
        for page in self._page_infos:
            size += page.size
        return size

    def add_page_info(self, v_addr, f_addr, size):
        if size <= 0:
            return

        new_page = MapPointerConverter.PageInfo(v_addr, f_addr, size)
        i = 0
        for page in self._page_infos:
            # insert the new page based on its virtual pointer
            if v_addr <= page.v_addr:
                break
            i += 1
        self._page_infos.insert(i, new_page)

    def v_ptr_to_f_ptr(self, ptr):
        '''Converts a virtual pointer in a map into a file pointer.'''
        for info in self._page_infos:
            if ptr in info.v_addr_range:
                return ptr - info.v_addr + info.f_addr
        return -0x7FffFFff

    def f_ptr_to_v_ptr(self, ptr):
        '''Converts a file pointer in a map into a virtual pointer.'''
        for info in self._page_infos:
            if ptr in info.f_addr_range:
                return ptr + info.v_addr - info.f_addr
        return -0x7FffFFff



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...hek.defs.bitm import *
from .objs.bitm import StubbsBitmTag

def get():
    return bitm_def

bitm_def = TagDef("bitm",
    blam_header('bitm', 7),
    bitm_body,

    ext=".bitmap", endian=">", tag_cls=StubbsBitmTag,
    subdefs = {'pixel_root':pixel_root}
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from ...common_descs import *
from .objs.tag import HekTag
from supyr_struct.defs.tag_def import TagDef

cdmg_body = Struct("tagdata",
    from_to_wu("radius"),
    float_zero_to_one("cutoff_scale"),
    Pad(24),

    QStruct("vibrate_parameters",
        float_zero_to_one("low_frequency"),
        float_zero_to_one("high_frequency"),
        Pad(24),
        ),

    Struct("camera_shaking",
        float_wu("random_translation"),
        float_rad("random_rotation"),  # radians
        Pad(12),

        SEnum16("wobble_function", *animation_functions),
        Pad(2),
        float_sec("wobble_function_period"),
        Float("wobble_weight"),
        Pad(192),
        ),

    Struct("damage",
        SEnum16("priority",
            "none",
            "harmless",
            {NAME: "backstab", GUI_NAME: "lethal to the unsuspecting"},
            "emp",
            ),
        SEnum16("category", *damage_category),
        Bool32("flags",
            "does_not_hurt_owner",
            {NAME: "headshot", GUI_NAME: "can cause headshots"},
            "pings_resistant_units",
            "does_not_hurt_friends",
            "does_not_ping_shields",
            "detonates_explosives",
            "only_hurts_shields",
            "causes_flaming_death",
            {NAME: "indicator_points_down",
             GUI_NAME: "damage indicator always points down"},
            "skips_shields",
            "only_hurts_one_infection_form",
            {NAME: "multiplayer_headshot",
             GUI_NAME: "can cause multiplayer headshots"},
            "infection_form_pop",
            ),
        Pad(4),
        Float("damage_lower_bound"),
        QStruct("damage_upper_bound", INCLUDE=from_to),
        float_zero_to_one("vehicle_passthrough_penalty"),
        Pad(4),
        float_zero_to_one("stun"),
        float_zero_to_one("maximum_stun"),
        float_sec("stun_time"),
        Pad(4),
        float_zero_to_inf("instantaneous_acceleration"),
        Pad(8),
        ),

    damage_modifiers,
    SIZE=512,
    )


def get():
    return cdmg_def

cdmg_def = TagDef("cdmg",
    blam_header('cdmg'),
    cdmg_body,

    ext=".continuous_damage_effect", endian=">", tag_cls=HekTag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: -DeToX-
# 	Mapped out the Raw Entry Table struct
# revision: 3		author: Lord Zedd
# 	Matching Zonesets
# revision: 4		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from .play import play_body
from supyr_struct.defs.tag_def import TagDef


zone_resource_type = Struct("resource_type",
    BytesRaw("guid", SIZE=16),
    Array("unknown_array", SUB_STRUCT=SInt16("unknown"), SIZE=4),
    h3_string_id("name"),
    ENDIAN=">", SIZE=28
    )


zone_resource_structure_type = Struct("resource_structure_type",
    BytesRaw("guid", SIZE=16),
    h3_string_id("name"),
    ENDIAN=">", SIZE=20
    )


zone_tag_resource_resource_fixup = Struct("resource_fixup",
    SInt32("block_offset"),
    h3_page_offset_info("address_info"),
    ENDIAN=">", SIZE=8,
    COMMENT="""
Block Offset: Offset in tag data sub-buffer.
Address: Fix-up address(either in the data sub-buffer, or a cache-resource offset."""
    )


zone_tag_resource_resource_definition_fixup = Struct("resource_definition_fixup",
    h3_page_offset_info("offset_info"),
    dyn_senum32("resource_structure_type",
        DYN_NAME_PATH="........resource_structure_types.STEPTREE[DYN_I].name.string",
        WIDGET_WIDTH=32
        ),
    ENDIAN=">", SIZE=8
    )


zone_tag_resource = Struct("tag_resource",
    h3_dependency("parent_tag"),
    UInt16("salt"),
    dyn_senum8("resource_type",
        DYN_NAME_PATH=".....resource_types.STEPTREE[DYN_I].name.string",
        WIDGET_WIDTH=32
        ),
    UInt8("flags"),
    SInt32("fixup_info_offset"),
    SInt32("fixup_info_size"),
    SInt32("secondary_fixup_info_offset"),
    Bool16("pages_used",
        "primary",
        "secondary",
        ),
    SInt16("segment_index"),
    h3_page_offset_info("root_definition_address_info"),
    h3_reflexive("resource_fixups", zone_tag_resource_resource_fixup),
    h3_reflexive("resource_definition_fixups", zone_tag_resource_resource_definition_fixup),
    ENDIAN=">", SIZE=64,
    )


zone_zoneset_raw_pool = Struct("raw_pool",
    Bool32("active_members", *("active_member_%s" % i for i in range(32))),
    ENDIAN=">", SIZE=4
    )


zone_zoneset_resource_type = QStruct("resource_type",
    UInt16("unknown0"),
    UInt16("unknown1"),
    UInt16("unknown2"),
    UInt16("unknown3"),
    UInt16("unknown4"),
    UInt16("unknown5"),
    UInt16("unknown6"),
    UInt16("unknown7"),
    UInt16("unknown8"),
    UInt16("unknown9"),
    #BytesRaw("unknown", SIZE=20, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=20
    )


zone_zoneset = Struct("zoneset",
    h3_reflexive("required_raw_pool", zone_zoneset_raw_pool),
    BytesRaw("unknown_0", SIZE=12, VISIBLE=False),
    h3_reflexive("optional_raw_pool", zone_zoneset_raw_pool),
    h3_reflexive("optional_raw_pool_2", zone_zoneset_raw_pool),
    BytesRaw("unknown_1", SIZE=20, VISIBLE=False),
    h3_string_id("set_name"),
    h3_reflexive("resource_types", zone_zoneset_resource_type),
    h3_reflexive("required_tag_pool", zone_zoneset_raw_pool),
    h3_reflexive("optional_tag_pool", zone_zoneset_raw_pool),
    BytesRaw("unknown_2", SIZE=12, VISIBLE=False),
    ENDIAN=">", SIZE=120
    )


zone_scenario_zoneset_group = Struct("scenario_zoneset_group",
    h3_string_id("name"),
    SInt32("bsp_group_index"),
    Bool32("import_loaded_bsps", *("bsp_%s" % i for i in range(32))),
    Bool32("loaded_bsps", *("bsp_%s" % i for i in range(32))),
    Bool32("loaded_designer_zonesets", *("set_%s" % i for i in range(32))),
    Bool32("unknown_loaded_designer_zonesets", *("set_%s" % i for i in range(32))),
    Bool32("unloaded_designer_zonesets", *("set_%s" % i for i in range(32))),
    Bool32("loaded_cinematic_zonesets", *("set_%s" % i for i in range(32))),
    SInt32("bsp_atlas_index"),
    ENDIAN=">", SIZE=36
    )


zone_scenario_bsp = Struct("scenario_bsp",
    h3_dependency("bsp"),
    ENDIAN=">", SIZE=16
    )


zone_unknown_4 = QStruct("unknown_4",
    UInt32("unknown0"),
    UInt32("unknown1"),
    UInt32("unknown2"),
    UInt32("unknown3"),
    UInt32("unknown4"),
    #BytesRaw("unknown", SIZE=20, VISIBLE=False),
    VISIBLE=False,
    ENDIAN=">", SIZE=20
    )


zone_prediction_a = QStruct("prediction_a",
    UInt32("key"),
    ENDIAN=">", SIZE=4
    )


zone_prediction_b = QStruct("prediction_b",
    SInt16("overall_index"),
    SInt16("a_count"),
    SInt32("a_index"),
    ENDIAN=">", SIZE=8
    )


zone_prediction_c = QStruct("prediction_c",
    SInt16("overall_index"),
    SInt16("b_index"),
    ENDIAN=">", SIZE=4
    )


zone_prediction_d_tag = QStruct("prediction_d_tag",
    SInt16("c_count"),
    SInt16("c_index"),
    SInt16("a_count"),
    SInt16("a_index"),
    ENDIAN=">", SIZE=8
    )


zone_prediction_d2_tag = Struct("prediction_d2_tag",
    dependency_uint32("tag", VISIBLE=False),
    SInt32("first_value"),
    SInt32("second_value"),
    ENDIAN=">", SIZE=12
    )


zone_body = Struct("tagdata",
    SEnum16("map_type", *zone_map_type),
    SInt16("flags"),
    h3_reflexive("resource_types", zone_resource_type,
        DYN_NAME_PATH='.name.string'),
    h3_reflexive("resource_structure_types", zone_resource_structure_type,
        DYN_NAME_PATH='.name.string'),
    Struct("play_data", INCLUDE=play_body),
    h3_reflexive("tag_resources", zone_tag_resource),
    Struct("zonesets",
        h3_reflexive("designer", zone_zoneset),
        h3_reflexive("global", zone_zoneset),
        BytesRaw("unknown_0", SIZE=12, VISIBLE=False),
        h3_reflexive("unattached", zone_zoneset),
        h3_reflexive("disc_forbidden", zone_zoneset),
        h3_reflexive("disc_always_streaming", zone_zoneset),
        h3_reflexive("bsp_1", zone_zoneset),
        h3_reflexive("bsp_2", zone_zoneset),
        h3_reflexive("bsp_3", zone_zoneset),
        h3_reflexive("cinematic", zone_zoneset),
        h3_reflexive("scenario", zone_zoneset),
        BytesRaw("unknown_1", SIZE=24, VISIBLE=False),
        ),
    h3_reflexive("scenario_zoneset_groups", zone_scenario_zoneset_group),
    h3_reflexive("scenario_bsps", zone_scenario_bsp),
    BytesRaw("unknown_2", SIZE=36, VISIBLE=False),
    h3_rawdata_ref("fixup_info"),
    BytesRaw("unknown_3", SIZE=20, VISIBLE=False),
    h3_reflexive("unknown_4", zone_unknown_4),
    BytesRaw("unknown_5", SIZE=96, VISIBLE=False),
    h3_reflexive("prediction_a", zone_prediction_a),
    h3_reflexive("prediction_b", zone_prediction_b),
    h3_reflexive("prediction_c", zone_prediction_c),
    h3_reflexive("prediction_d_tags", zone_prediction_d_tag),
    h3_reflexive("prediction_d2_tags", zone_prediction_d2_tag),
    SInt32("campaign_id"),
    SInt32("map_id"),
    ENDIAN=">", SIZE=532
    )


def get():
    return zone_def

zone_def = TagDef("zone",
    h3_blam_header('zone'),
    zone_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["zone"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: DarkShallFall
# 	Labeled most of the tag
# revision: 3		author: DarkShallFall
# 	H2
# revision: 4		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef

clwd_integration_type = (
    "verlet",
    )


clwd_vertice = Struct("vertice",
    QStruct("initial_position", INCLUDE=xyz_float),
    QStruct("uv", INCLUDE=ij_float),
    ENDIAN=">", SIZE=20
    )


clwd_indice = Struct("indice",
    SInt16("idx"),
    ENDIAN=">", SIZE=2
    )


clwd_link = Struct("link",
    SInt16("index_1"),
    SInt16("index_2"),
    Float("default_distance"),
    ENDIAN=">", SIZE=8
    )


clwd_body = Struct("tagdata",
    Bool32("flags",
        "doesn_t_use_wind",
        "uses_grid_attach_top",
        ),
    h3_string_id("marker_attachment_name"),
    h3_string_id("second_marker_attachment_name"),
    h3_dependency("shader"),
    SInt16("grid_x_dimension"),
    SInt16("grid_y_dimension"),
    QStruct("grid_spacing", INCLUDE=xy_float),
    BytesRaw("unknown_0", SIZE=12, VISIBLE=False),
    SEnum16("integration_type", *clwd_integration_type),
    SInt16("number_iterations"),
    Float("weight"),
    Float("drag"),
    Float("wind_scale"),
    Float("wind_flappiness_scale"),
    Float("longest_rod"),
    BytesRaw("unknown_1", SIZE=24, VISIBLE=False),
    h3_reflexive("vertices", clwd_vertice),
    h3_reflexive("indices", clwd_indice),
    BytesRaw("unknown_2", SIZE=12, VISIBLE=False),
    h3_reflexive("links", clwd_link),
    ENDIAN=">", SIZE=148
    )


def get():
    return clwd_def

clwd_def = TagDef("clwd",
    h3_blam_header('clwd'),
    clwd_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["clwd"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

############# Credits and version info #############
# Definition generated from Assembly XML tag def
#	 Date generated: 2018/12/03  04:56
#
# revision: 1		author: Assembly
# 	Generated plugin from scratch.
# revision: 2		author: Lord Zedd
# 	Basically done.
# revision: 3		author: Moses_of_Egypt
# 	Cleaned up and converted to SuPyr definition
#
####################################################

from ..common_descs import *
from .objs.tag import *
from supyr_struct.defs.tag_def import TagDef


wrot_animation_definition = Struct("animation_definition",
    UInt32("frame"),
    SEnum16("anchor", *wscl_animation_definition_anchor),
    SInt16("unknown_0", VISIBLE=False),
    QStruct("custom_anchor", INCLUDE=xy_float),
    Float("rotation_amount"),
    BytesRaw("unknown_1", SIZE=12, VISIBLE=False),
    ENDIAN=">", SIZE=32
    )


wrot_body = Struct("tagdata",
    Bool32("animation_flags",
        "loops",
        ),
    h3_reflexive("animation_definition", wrot_animation_definition),
    h3_rawdata_ref("date"),
    ENDIAN=">", SIZE=36
    )


def get():
    return wrot_def

wrot_def = TagDef("wrot",
    h3_blam_header('wrot'),
    wrot_body,

    ext=".%s" % h3_tag_class_fcc_to_ext["wrot"], endian=">", tag_cls=H3Tag
    )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import os

from copy import deepcopy
from math import pi, log
from pathlib import Path
from struct import unpack, pack_into
from traceback import format_exc
from types import MethodType

from supyr_struct.buffer import BytearrayBuffer
from supyr_struct.field_types import FieldType
from supyr_struct.defs.frozen_dict import FrozenDict

from reclaimer.halo_script.hsc import get_hsc_data_block,\
     HSC_IS_SCRIPT_OR_GLOBAL, SCRIPT_OBJECT_TYPES_TO_SCENARIO_REFLEXIVES
from reclaimer.common_descs import make_dependency_os_block
from reclaimer.hek.defs.snd_ import snd__meta_stub_blockdef
from reclaimer.hek.defs.sbsp import sbsp_meta_header_def
from reclaimer.os_hek.defs.gelc    import gelc_def
from reclaimer.os_v4_hek.defs.coll import fast_coll_def
from reclaimer.os_v4_hek.defs.sbsp import fast_sbsp_def
from reclaimer.os_v4_hek.handler   import OsV4HaloHandler
from reclaimer.meta.wrappers.byteswapping import raw_block_def, byteswap_animation,\
     byteswap_uncomp_verts, byteswap_comp_verts, byteswap_tris,\
     byteswap_coll_bsp, byteswap_sbsp_meta, byteswap_scnr_script_syntax_data,\
     byteswap_pcm16_samples
from reclaimer.meta.wrappers.halo_map import HaloMap
from reclaimer.meta.wrappers.halo1_rsrc_map import Halo1RsrcMap, inject_sound_data, get_is_xbox_map
from reclaimer.meta.wrappers.map_pointer_converter import MapPointerConverter
from reclaimer.meta.wrappers.tag_index_manager import TagIndexManager
from reclaimer import data_extraction
from reclaimer.constants import tag_class_fcc_to_ext
from reclaimer.util.compression import compress_normal32, decompress_normal32
from reclaimer.util import is_overlapping_ranges, is_valid_ascii_name_str,\
     int_to_fourcc

from supyr_struct.util import is_path_empty


__all__ = ("Halo1Map", "Halo1RsrcMap")


def reparse_reflexive(block, new_size, pointer_converter,
                      map_data, tag_index_manager):
    steptree = block.STEPTREE
    file_ptr = pointer_converter.v_ptr_to_f_ptr(block.pointer)

    if (file_ptr + new_size * steptree.get_desc("SIZE", 0) > len(map_data) or
        file_ptr <= 0):
        return

    # read the palette array from the map
    with FieldType.force_little:
        del steptree[:]
        steptree.extend(new_size)
        steptree.TYPE.parser(
            steptree.desc, node=steptree,
            map_pointer_converter=pointer_converter,
            rawdata=map_data, offset=file_ptr,
            safe_mode=False, parsing_resource=False,
            tag_index_manager=tag_index_manager)


class Halo1Map(HaloMap):
    '''Generation 1 map'''
    ce_rsrc_sound_indexes_by_path = None
    ce_tag_indexs_by_paths = None
    sound_rsrc_id = None
    defs = None

    tag_defs_module = "reclaimer.os_v4_hek.defs"
    tag_classes_to_load = tuple(sorted(tag_class_fcc_to_ext.keys()))

    handler_class = OsV4HaloHandler

    force_checksum = False

    inject_rawdata = Halo1RsrcMap.inject_rawdata

    bsp_magics  = ()
    bsp_sizes   = ()
    bsp_headers = ()
    bsp_header_offsets = ()
    bsp_pointer_converters = ()

    data_extractors = data_extraction.h1_data_extractors

    def __init__(self, maps=None):
        HaloMap.__init__(self, maps)

        self.resource_map_class = Halo1RsrcMap
        self.ce_rsrc_sound_indexes_by_path = {}
        self.ce_tag_indexs_by_paths = {}

        self.bsp_magics = {}
        self.bsp_sizes  = {}
        self.bsp_header_offsets = {}
        self.bsp_headers = {}
        self.bsp_pointer_converters = {}

        self.setup_tag_headers()

    @property
    def decomp_file_ext(self):
        if self.engine == "halo1yelo":
            return ".yelo"
        elif self.engine == "halo1vap":
            return ".vap"
        else:
            return ".map"

    def setup_defs(self):
        this_class = type(self)
        if this_class.defs is None:
            this_class.defs = defs = {}
            print("    Loading definitions in '%s'" % self.tag_defs_module)
            this_class.handler = self.handler_class(
                build_reflexive_cache=False, build_raw_data_cache=False,
                debug=2)

            this_class.defs = dict(this_class.handler.defs)
            this_class.defs["coll"] = fast_coll_def
            this_class.defs["gelc"] = gelc_def
            this_class.defs["sbsp"] = fast_sbsp_def
            this_class.defs = FrozenDict(this_class.defs)

        # make a shallow copy for this instance to manipulate
        self.defs = dict(self.defs)

    def ensure_sound_maps_valid(self):
        sounds = self.maps.get("sounds")
        if not sounds or self.is_resource:
            return

        if id(sounds) == self.sound_rsrc_id and (
                self.ce_rsrc_sound_indexes_by_path and
                self.ce_tag_indexs_by_paths):
            return

        self.sound_rsrc_id = id(sounds)
        if self.engine in ("halo1ce", "halo1yelo", "halo1vap"):
            # ce resource sounds are recognized by tag_path
            # so we must cache their offsets by their paths
            rsrc_snd_map = self.ce_rsrc_sound_indexes_by_path = {}
            inv_snd_map  = self.ce_tag_indexs_by_paths = {}

            if sounds is not None:
                i = 0
                for tag_header in sounds.rsrc_map.data.tags:
                    rsrc_snd_map[tag_header.tag.path] = i
                    i += 1

            i = 0
            for tag_header in self.tag_index.tag_index:
                inv_snd_map[tag_header.path] = i
                i += 1

    def get_dependencies(self, meta, tag_id, tag_cls):
        tag_index_array = self.tag_index.tag_index
        if not self.is_indexed(tag_id):
            # not indexed. get dependencies like normal
            pass
        elif tag_cls != "snd!":
            # among the indexable tags, only sounds can have valid dependencies
            return ()
        elif not hasattr(meta, "pitch_ranges"):
            # tag is indexed AND we were provided with the indexed version
            rsrc_id = meta.promotion_sound.id & 0xFFff
            if rsrc_id == 0xFFFF: return ()

            sounds = self.maps.get("sounds")
            if   sounds is None: return ()
            elif rsrc_id >= len(sounds.tag_index.tag_index): return ()

            tag_path = sounds.tag_index.tag_index[rsrc_id].path
            inv_snd_map = getattr(self, 'ce_tag_indexs_by_paths', {})
            tag_id = inv_snd_map.get(tag_path, 0xFFFF)
            if tag_id >= len(tag_index_array): return ()

            ref = deepcopy(meta.promotion_sound)
            tag_index_ref = tag_index_array[tag_id]
            ref.tag_class.data = tag_index_ref.class_1.data
            ref.id             = tag_index_ref.id
            ref.filepath       = tag_index_ref.path

            return [ref]

        if self.handler is None: return ()

        dependency_cache = self.handler.tag_ref_cache.get(tag_cls)
        if not dependency_cache: return ()

        nodes = self.handler.get_nodes_by_paths(dependency_cache, (None, meta))
        dependencies = []

        for node in nodes:
            if node.id & 0xFFff == 0xFFFF:
                continue
            dependencies.append(node)

        if tag_cls == "scnr":
            # collect the tag references from the scenarios syntax data
            try:
                seen_tag_ids = set()
                syntax_data = get_hsc_data_block(meta.script_syntax_data.data)
                for node in syntax_data.nodes:
                    if (node.flags & HSC_IS_SCRIPT_OR_GLOBAL or
                        node.type not in range(24, 32)):
                        # not a tag index ref
                        continue

                    tag_index_id = node.data & 0xFFff
                    if (tag_index_id in range(len(tag_index_array)) and
                        tag_index_id not in seen_tag_ids):
                        seen_tag_ids.add(tag_index_id)
                        tag_index_ref = tag_index_array[tag_index_id]
                        dependencies.append(make_dependency_os_block(
                            tag_index_ref.class_1.enum_name, tag_index_ref.id,
                            tag_index_ref.path, tag_index_ref.path_offset))
            except Exception:
                pass

        return dependencies

    def setup_sbsp_pointer_converters(self):
        # get the scenario meta
        if self.scnr_meta is None:
            print("Cannot setup sbsp pointer converters without scenario tag.")
            return

        try:
            metadata_range = range(self.map_header.tag_index_header_offset,
                                   self.map_header.tag_index_header_offset +
                                   self.map_header.tag_data_size)
            invalid_bsp_tag_ids = [self.tag_index.scenario_tag_id & 0xFFff]
            i = 0
            for b in self.scnr_meta.structure_bsps.STEPTREE:
                bsp_id = b.structure_bsp.id & 0xFFff

                # these checks are necessary because apparently these structs
                # can be super fucked up, and might not affect anything
                if (bsp_id not in range(len(self.tag_index.tag_index)) or
                    bsp_id in invalid_bsp_tag_ids):
                    print("Scenario structure_bsp %s contains invalid tag_id." % i)
                elif is_overlapping_ranges(
                        metadata_range, range(b.bsp_pointer, b.bsp_size)):
                    print("Scenario structure_bsp %s contains invalid pointer." % i)
                elif self.tag_index.tag_index[bsp_id].id != b.structure_bsp.id:
                    print("Scenario structure_bsp %s contains invalid tag_id." % i)
                else:
                    self.bsp_header_offsets[bsp_id] = b.bsp_pointer
                    self.bsp_magics[bsp_id]         = b.bsp_magic
                    self.bsp_sizes[bsp_id]          = b.bsp_size

                    self.bsp_pointer_converters[bsp_id] = MapPointerConverter(
                        (b.bsp_magic, b.bsp_pointer, b.bsp_size)
                        )

                i += 1

            # read the sbsp headers
            for tag_id, offset in self.bsp_header_offsets.items():
                if self.engine == "halo1anni":
                    with FieldType.force_big:
                        header = sbsp_meta_header_def.build(
                            rawdata=self.map_data, offset=offset)
                else:
                    header = sbsp_meta_header_def.build(
                        rawdata=self.map_data, offset=offset)

                if header.sig != header.get_desc("DEFAULT", "sig"):
                    print("Sbsp header is invalid for '%s'" %
                          self.tag_index.tag_index[tag_id].path)
                self.bsp_headers[tag_id] = header
                self.tag_index.tag_index[tag_id].meta_offset = header.meta_pointer

        except Exception:
            print(format_exc())

    def load_map(self, map_path, **kwargs):
        HaloMap.load_map(self, map_path, **kwargs)

        tag_index = self.tag_index
        tag_index_array = tag_index.tag_index

        # cache the original paths BEFORE running basic deprotection
        self.cache_original_tag_paths()

        # make all contents of the map parseable
        self.basic_deprotection()

        self.tag_index_manager = TagIndexManager(tag_index_array)

        # add the tag data section
        self.map_pointer_converter.add_page_info(
            self.index_magic, self.map_header.tag_index_header_offset,
            self.map_header.tag_data_size
            )

        # cache the scenario meta
        try:
            self.scnr_meta = self.get_meta(self.tag_index.scenario_tag_id)
            if self.scnr_meta is None:
                print("Could not read scenario tag")
        except Exception:
            print(format_exc())
            print("Could not read scenario tag")

        self.setup_sbsp_pointer_converters()

        last_bsp_end = 0
        # calculate the start of the rawdata section
        for tag_id in self.bsp_headers:
            bsp_end = self.bsp_header_offsets[tag_id] + self.bsp_sizes[tag_id]
            if last_bsp_end < bsp_end:
                last_bsp_end = bsp_end

        # add the rawdata section
        self.map_pointer_converter.add_page_info(
            last_bsp_end, last_bsp_end,
            tag_index.model_data_offset - last_bsp_end,
            )

        # add the model data section
        if tag_index.SIZE == 40:
            # PC tag index
            self.map_pointer_converter.add_page_info(
                0, tag_index.model_data_offset,
                tag_index.model_data_size,
                )
        else:
            # XBOX tag index
            self.map_pointer_converter.add_page_info(
                0, tag_index.model_data_offset,
                (self.map_header.tag_index_header_offset -
                 tag_index.model_data_offset),
                )

        # get the globals meta
        try:
            matg_id = None
            for b in tag_index_array:
                if int_to_fourcc(b.class_1.data) == "matg":
                    matg_id = b.id & 0xFFff
                    break

            self.matg_meta = self.get_meta(matg_id)
            if self.matg_meta is None:
                print("Could not read globals tag")
        except Exception:
            print(format_exc())
            print("Could not read globals tag")

        if self.map_name == "sounds":
            for halo_map in self.maps.values():
                if hasattr(halo_map, "ensure_sound_maps_valid"):
                    halo_map.ensure_sound_maps_valid()

        self.clear_map_cache()

    def get_meta(self, tag_id, reextract=False, ignore_rsrc_sounds=False, **kw):
        '''
        Takes a tag reference id as the sole argument.
        Returns that tags meta data as a parsed block.
        '''
        if tag_id is None:
            return

        magic     = self.map_magic
        engine    = self.engine
        map_data  = self.map_data
        tag_index = self.tag_index
        tag_index_array = tag_index.tag_index

        # if we are given a 32bit tag id, mask it off
        tag_id &= 0xFFFF
        tag_index_ref = self.tag_index_manager.get_tag_index_ref(tag_id)
        if tag_index_ref is None:
            return

        tag_cls = None
        if tag_id == (tag_index.scenario_tag_id & 0xFFFF):
            tag_cls = "scnr"
        elif tag_index_ref.class_1.enum_name not in ("<INVALID>", "NONE"):
            tag_cls = int_to_fourcc(tag_index_ref.class_1.data)

        # if we dont have a defintion for this tag_cls, then return nothing
        if self.get_meta_descriptor(tag_cls) is None:
            return

        self.ensure_sound_maps_valid()
        pointer_converter = self.bsp_pointer_converters.get(
            tag_id, self.map_pointer_converter)
        offset = tag_index_ref.meta_offset

        if tag_cls is None:
            # couldn't determine the tag class
            return
        elif self.is_indexed(tag_id) and (
                tag_cls != "snd!" or not ignore_rsrc_sounds):
            # tag exists in a resource cache
            tag_id = offset

            rsrc_map = None
            if tag_cls == "snd!" and "sounds" in self.maps:
                rsrc_map = self.maps["sounds"]
                sound_mapping = self.ce_rsrc_sound_indexes_by_path
                tag_path = tag_index_ref.path
                if sound_mapping is None or tag_path not in sound_mapping:
                    return

                tag_id = sound_mapping[tag_path]//2

            elif tag_cls == "bitm" and "bitmaps" in self.maps:
                rsrc_map = self.maps["bitmaps"]
                tag_id = tag_id//2

            elif "loc" in self.maps:
                rsrc_map = self.maps["loc"]
                # this resource tag COULD be in a yelo loc.map, which
                # means we will need to set its tag class to what this
                # map specifies it as or else the resource map wont
                # know what type of tag to extract it as.
                rsrc_map.tag_index.tag_index[tag_id].class_1.set_to(
                    tag_index_ref.class_1.enum_name)

            if rsrc_map is None:
                return

            meta = rsrc_map.get_meta(tag_id)
            if tag_cls == "snd!":
                # since we're reading the resource tag from the perspective of
                # the map referencing it, we have more accurate information
                # about which other sound it could be referencing. This is only
                # a concern when dealing with open sauce resource maps, as they
                # could have additional promotion sounds we cant statically map
                try:
                    # read the meta data from the map
                    with FieldType.force_little:
                        snd_stub = snd__meta_stub_blockdef.build(
                            rawdata=map_data,
                            offset=pointer_converter.v_ptr_to_f_ptr(offset),
                            tag_index_manager=self.tag_index_manager)
                    meta.promotion_sound = snd_stub.promotion_sound
                except Exception:
                    print(format_exc())

            return meta
        elif not reextract:
            if tag_id == tag_index.scenario_tag_id & 0xFFff and self.scnr_meta:
                return self.scnr_meta
            elif tag_cls == "matg" and self.matg_meta:
                return self.matg_meta

        force_parsing_rsrc = False
        if tag_cls in ("antr", "magy") and not kw.get("disable_tag_cleaning"):
            force_parsing_rsrc = True

        desc = self.get_meta_descriptor(tag_cls)
        block = [None]
        try:
            # read the meta data from the map
            with FieldType.force_little:
                desc['TYPE'].parser(
                    desc, parent=block, attr_index=0, rawdata=map_data,
                    map_pointer_converter=pointer_converter,
                    offset=pointer_converter.v_ptr_to_f_ptr(offset),
                    tag_index_manager=self.tag_index_manager,
                    safe_mode=not kw.get("disable_safe_mode"),
                    parsing_resource=force_parsing_rsrc)
        except Exception:
            print(format_exc())
            if not kw.get("allow_corrupt"):
                return

        meta = block[0]
        try:
            if tag_cls == "bitm" and get_is_xbox_map(engine):
                for bitmap in meta.bitmaps.STEPTREE:
                    # make sure to set this for all xbox bitmaps
                    # so they can be interpreted properly
                    bitmap.base_address = 1073751810

            self.record_map_cache_read(tag_id, 0)
            if self.map_cache_over_limit():
                self.clear_map_cache()

            if not kw.get("ignore_rawdata", False):
                self.inject_rawdata(meta, tag_cls, tag_index_ref)
        except Exception:
            print(format_exc())
            if not kw.get("allow_corrupt"):
                meta = None

        if not kw.get("disable_tag_cleaning"):
            try:
                self.clean_tag_meta(meta, tag_id, tag_cls)
            except Exception:
                print(format_exc())

        return meta

    def clean_tag_meta(self, meta, tag_id, tag_cls):
        tag_index_array = self.tag_index.tag_index

        if tag_cls in ("antr", "magy"):
            highest_valid = -1
            found_valid = False
            animations = meta.animations.STEPTREE
            main_node_count = animations[0].node_count if animations else 0
            main_node_list_checksum = animations[0].node_list_checksum if animations else 0

            permutation_chains = {}
            for i in range(len(animations)):
                if i in permutation_chains:
                    continue

                permutation_chains[i] = i
                next_anim = animations[i].next_animation
                while (next_anim in range(len(animations)) and
                       next_anim not in permutation_chains):
                    permutation_chains[next_anim] = i
                    next_anim = animations[next_anim].next_animation

            anims_to_remove = []

            for i in range(len(animations)):
                if self.engine != "halo1yelo" and i >= 256:
                    # cap it to the non-OS limit of 256 animations
                    break

                anim = animations[i]
                valid = is_valid_ascii_name_str(anim.name)

                trans_int = anim.trans_flags0 + (anim.trans_flags1 << 32)
                rot_int   = anim.rot_flags0   + (anim.rot_flags1   << 32)
                scale_int = anim.scale_flags0 + (anim.scale_flags1 << 32)

                trans_flags = (bool(trans_int & (1 << i))
                               for i in range(anim.node_count))
                rot_flags   = (bool(rot_int   & (1 << i))
                               for i in range(anim.node_count))
                scale_flags = (bool(scale_int & (1 << i))
                               for i in range(anim.node_count))

                expected_frame_size = (12 * sum(trans_flags) +
                                       8  * sum(rot_flags) +
                                       4  * sum(scale_flags))
                expected_frame_info_size = {1: 8, 2: 12, 3: 16}.get(
                    anim.frame_info_type.data, 0) * anim.frame_count
                expected_frame_data_size = expected_frame_size * anim.frame_count
                expected_default_data_size = (
                    anim.node_count * (12 + 8 + 4) - anim.frame_size)

                if anim.frame_count == 0:
                    expected_default_data_size = 0

                if (anim.type.enum_name == "<INVALID>" or
                    anim.frame_info_type.enum_name == "<INVALID>"):
                    valid = False
                elif (anim.node_count != main_node_count or
                      anim.node_count not in range(1, 65)):
                    valid = False
                elif anim.first_permutation_index != permutation_chains[i]:
                    valid = False
                elif not anim.flags.compressed_data:
                    if anim.default_data.size < expected_default_data_size:
                        valid = False
                    elif anim.frame_info.size < expected_frame_info_size:
                        valid = False
                    elif anim.frame_data.size < expected_frame_data_size:
                        valid = False
                    elif anim.frame_size != expected_frame_size:
                        valid = False
                elif anim.offset_to_compressed_data >= anim.frame_data.size:
                    valid = False

                if valid:
                    highest_valid = i
                    if not found_valid:
                        main_node_count = anim.node_count
                        main_node_list_checksum = anim.node_list_checksum
                        found_valid = True
                else:
                    # delete the animation info
                    anims_to_remove.append(i)

            # make sure all animations have the same node count, checksum, and a name
            for i in anims_to_remove:
                animations.pop(i)
                animations.insert(i)
                animations[i].node_count = main_node_count
                animations[i].node_list_checksum = main_node_list_checksum
                animations[i].name = "REMOVED_%s" % i

            # remove the highest invalid animations
            if highest_valid + 1 < len(animations):
                del animations[highest_valid + 1: ]

            # inject the animation data for all remaining animations since
            # it's not safe to try and read it all at parse time
            for anim in animations:
                for block in (anim.default_data, anim.frame_info, anim.frame_data):
                    if not block.size:
                        continue

                    file_ptr = self.map_pointer_converter.v_ptr_to_f_ptr(
                        block.pointer)
                    if not block.pointer or file_ptr < 0:
                        file_ptr = block.raw_pointer

                    if file_ptr + block.size > len(self.map_data) or file_ptr <= 0:
                        continue

                    try:
                        self.map_data.seek(file_ptr)
                        block.data = bytearray(self.map_data.read(block.size))
                    except Exception:
                        print("Couldn't read animation data.")

        elif tag_cls in ("sbsp", "coll"):
            if tag_cls == "sbsp" :
                bsps = meta.collision_bsp.STEPTREE
            else:
                bsps = []
                for node in meta.nodes.STEPTREE:
                    bsps.extend(node.bsps.STEPTREE)

            for bsp in bsps:
                highest_used_vert = -1
                edge_data = bsp.edges.STEPTREE
                vert_data = bsp.vertices.STEPTREE
                for i in range(0, len(edge_data), 24):
                    v0_i = (edge_data[i] +
                            (edge_data[i + 1] << 8) +
                            (edge_data[i + 2] << 16) +
                            (edge_data[i + 3] << 24))
                    v1_i = (edge_data[i + 4] +
                            (edge_data[i + 5] << 8) +
                            (edge_data[i + 6] << 16) +
                            (edge_data[i + 7] << 24))
                    highest_used_vert = max(highest_used_vert, v0_i, v1_i)

                if highest_used_vert * 16 < len(vert_data):
                    del vert_data[(highest_used_vert + 1) * 16: ]
                    bsp.vertices.size = highest_used_vert + 1

        elif tag_cls in ("mode", "mod2"):
            used_shaders = set()
            shaders = meta.shaders.STEPTREE

            for geom in meta.geometries.STEPTREE:
                for part in geom.parts.STEPTREE:
                    if part.shader_index >= 0:
                        used_shaders.add(part.shader_index)

            # determine the max number of shader indices actually used
            # by all the geometry parts, and reparse them with that many.
            if used_shaders:
                try:
                    reparse_reflexive(
                        meta.shaders, max(used_shaders) + 1,
                        self.map_pointer_converter,
                        self.map_data, self.tag_index_manager)
                except Exception:
                    print(format_exc())
                    print("Couldn't re-parse %s data." % meta.shaders)

            new_i = 0
            rebase_map = {}
            new_shaders = [None] * len(used_shaders)
            for i in sorted(used_shaders):
                new_shaders[new_i] = shaders[i]
                rebase_map[i] = new_i
                new_i += 1

            # rebase the shader indices
            for geom in meta.geometries.STEPTREE:
                for part in geom.parts.STEPTREE:
                    if part.shader_index in range(len(shaders)):
                        part.shader_index = rebase_map[part.shader_index]
                    else:
                        part.shader_index = -1

            shaders[:] = new_shaders

        elif tag_cls == "scnr":
            skies = meta.skies.STEPTREE
            comments = meta.comments.STEPTREE

            highest_valid_sky = -1
            for i in range(len(skies)):
                sky = skies[i].sky
                sky_tag_index_id = sky.id & 0xFFff
                if (sky_tag_index_id not in range(len(tag_index_array)) or
                    tag_index_array[sky_tag_index_id].id != sky.id):
                    # invalid sky
                    sky.id = 0xFFffFFff
                    sky.filepath = ""
                    sky.tag_class.set_to("sky")
                else:
                    highest_valid_sky = i

            # clear the highest invalid skies
            del skies[highest_valid_sky + 1: ]

            # clear the child scenarios since they aren't used
            del meta.child_scenarios.STEPTREE[:]

            # determine if there are any fucked up comments
            comments_to_keep = set()
            for i in range(len(comments)):
                comment = comments[i]
                if max(max(comment.position), abs(min(comment.position))) > 5000:
                    # check if the position is outside halos max world bounds
                    continue

                if not (comment.comment_data.data and
                        is_valid_ascii_name_str(comment.comment_data.data)):
                    comments_to_keep.add(i)

            if len(comments_to_keep) != len(comments):
                # clean up any fucked up comments
                comments[:] = [comments[i] for i in sorted(comments_to_keep)]

            syntax_data = get_hsc_data_block(meta.script_syntax_data.data)
            script_nodes_modified = False

            # clean up any fucked up palettes
            for pal_block, inst_block in (
                    (meta.sceneries_palette, meta.sceneries),
                    (meta.bipeds_palette, meta.bipeds),
                    (meta.vehicles_palette, meta.vehicles),
                    (meta.equipments_palette, meta.equipments),
                    (meta.weapons_palette, meta.weapons),
                    (meta.machines_palette, meta.machines),
                    (meta.controls_palette, meta.controls),
                    (meta.light_fixtures_palette, meta.light_fixtures),
                    (meta.sound_sceneries_palette, meta.sound_sceneries),
                    ):
                palette, instances = pal_block.STEPTREE, inst_block.STEPTREE

                used_pal_indices = set(inst.type for inst in instances
                                       if inst.type >= 0)
                script_nodes_to_modify = set()

                if inst_block.NAME == "bipeds":
                    # determine which palette indices are used by script data
                    for i in range(len(syntax_data.nodes)):
                        node = syntax_data.nodes[i]
                        # 35 == "actor_type"  script type
                        if node.type == 35 and not(node.flags & HSC_IS_SCRIPT_OR_GLOBAL):
                            script_nodes_to_modify.add(i)
                            used_pal_indices.add(node.data & 0xFFff)

                # determine the max number of palette indices actually used by all
                # the object instances, and reparse the palette with that many.
                if used_pal_indices:
                    try:
                        reparse_reflexive(
                            pal_block, max(used_pal_indices) + 1,
                            self.map_pointer_converter,
                            self.map_data, self.tag_index_manager)
                    except Exception:
                        print(format_exc())
                        print("Couldn't re-parse %s data." % pal_block.NAME)

                # figure out what to rebase the palette swatch indices to
                new_i = 0
                rebase_map = {}
                new_palette = [None] * len(used_pal_indices)
                for i in sorted(used_pal_indices):
                    new_palette[new_i] = palette[i]
                    rebase_map[i] = new_i
                    new_i += 1

                # rebase the palette indices of the instances and
                # move the palette swatches into their new indices
                for inst in instances:
                    if inst.type in range(len(instances)):
                        inst.type = rebase_map[inst.type]
                    else:
                        inst.type = -1

                palette[:] = new_palette

                # modify the script syntax nodes that need to be
                for i in script_nodes_to_modify:
                    node = syntax_data.nodes[i]
                    salt = node.data & 0xFFff0000
                    cur_index = node.data & 0xFFff
                    new_index = rebase_map[cur_index]
                    if cur_index != new_index:
                        script_nodes_modified = True
                        node.data = salt + new_index

            # replace the script syntax data
            if script_nodes_modified:
                with FieldType.force_little:
                    new_script_data = syntax_data.serialize()
                    meta.script_syntax_data.data[: len(new_script_data)] = new_script_data
                    meta.script_syntax_data.size = len(meta.script_syntax_data.data)

        elif tag_cls in ("tagc", "Soul"):
            tag_collection = meta[0].STEPTREE
            highest_valid = -1
            for i in range(len(tag_collection)):
                tag_ref = tag_collection[i][0]
                if tag_cls == "Soul" and tag_ref.tag_class.enum_name != "ui_widget_definition":
                    continue
                elif (tag_ref.id & 0xFFff) not in range(len(tag_index_array)):
                    continue
                elif tag_index_array[tag_ref.id & 0xFFff].id != tag_ref.id:
                    continue

                highest_valid = i

            # clear the highest invalid entries
            del tag_collection[highest_valid + 1: ]

    def meta_to_tag_data(self, meta, tag_cls, tag_index_ref, **kwargs):
        magic      = self.map_magic
        engine     = self.engine
        map_data   = self.map_data
        tag_index  = self.tag_index
        byteswap = kwargs.get("byteswap", True)

        predicted_resources = []

        if hasattr(meta, "obje_attrs"):
            predicted_resources.append(meta.obje_attrs.predicted_resources)

            # fix the change colors permutations
            for change_color in meta.obje_attrs.change_colors.STEPTREE:
                cutoff = 0
                for perm in change_color.permutations.STEPTREE:
                    perm.weight, cutoff = perm.weight - cutoff, perm.weight


        if tag_cls == "actv":
            # multiply grenade velocity by 30
            meta.grenades.grenade_velocity *= 30

        elif tag_cls in ("antr", "magy"):
            # byteswap animation data
            for anim in meta.animations.STEPTREE:
                if not byteswap: break
                byteswap_animation(anim)

        elif tag_cls == "bitm":
            # set the size of the compressed plate data to nothing
            meta.compressed_color_plate_data.STEPTREE = BytearrayBuffer()

            # to enable compatibility with my bitmap converter we'll set the
            # base address to a certain constant based on the console platform
            is_xbox = get_is_xbox_map(engine)

            new_pixels_offset = 0

            # uncheck the prefer_low_detail flag and
            # set up the pixels_offset correctly.
            for bitmap in meta.bitmaps.STEPTREE:
                bitmap.flags.prefer_low_detail = is_xbox
                bitmap.pixels_offset = new_pixels_offset
                new_pixels_offset += bitmap.pixels_meta_size

                # clear some meta-only fields
                bitmap.pixels_meta_size = 0
                bitmap.bitmap_id_unknown1 = bitmap.bitmap_id_unknown2 = 0
                bitmap.bitmap_data_pointer = 0

                if is_xbox:
                    bitmap.base_address = 1073751810
                    if "dxt" in bitmap.format.enum_name:
                        # need to correct mipmap count on xbox dxt bitmaps.
                        # the game seems to prune the mipmap texels for any
                        # mipmaps whose dimensions are 2x2 or smaller

                        max_dim = max(bitmap.width, bitmap.height)
                        if 2 ** bitmap.mipmaps > max_dim:
                            # make sure the mipmap level isnt higher than the
                            # number of mipmaps that should be able to exist.
                            bitmap.mipmaps = int(log(max_dim, 2))

                        last_mip_dim = max_dim // (2 ** bitmap.mipmaps)
                        if last_mip_dim == 1:
                            bitmap.mipmaps -= 2
                        elif last_mip_dim == 2:
                            bitmap.mipmaps -= 1

                        if bitmap.mipmaps < 0:
                            bitmap.mipmaps = 0
                else:
                    bitmap.base_address = 0

        elif tag_cls == "cdmg":
            # divide camera shaking wobble period by 30
            meta.camera_shaking.wobble_function_period /= 30

        elif tag_cls == "coll":
            # byteswap the raw bsp collision data
            for node in meta.nodes.STEPTREE:
                for perm_bsp in node.bsps.STEPTREE:
                    if not byteswap: break
                    byteswap_coll_bsp(perm_bsp)

        elif tag_cls == "effe":
            # mask away the meta-only flags
            meta.flags.data &= 3
            for event in meta.events.STEPTREE:
                # tool exceptions if any parts reference a damage effect
                # tag type, but have an empty filepath for the reference
                parts = event.parts.STEPTREE
                for i in range(len(parts) - 1, -1, -1):
                    part = parts[i]
                    if (part.type.tag_class.enum_name == "damage_effect" and
                        not part.type.filepath):
                        parts.pop(i)

        elif tag_cls == "jpt!":
            # camera shaking wobble period by 30
            meta.camera_shaking.wobble_function_period /= 30

        elif tag_cls == "glw!":
            # increment enumerators properly
            for b in (meta.particle_rotational_velocity,
                      meta.effect_rotational_velocity,
                      meta.effect_translational_velocity,
                      meta.particle_distance_to_object,
                      meta.particle_size,
                      meta.particle_color):
                b.attachment.data += 1

        elif tag_cls == "lens":
            # DON'T multiply corona rotation by pi/180
            # reminder that this is not supposed to be changed
            pass # meta.corona_rotation.function_scale *= pi/180

        elif tag_cls == "ligh":
            # divide light time by 30
            meta.effect_parameters.duration /= 30

        elif tag_cls == "matg":
            # tool will fail to compile any maps if the
            # multiplayer_info or falling_damage is blank

            # make sure there is multiplayer info.
            multiplayer_info = meta.multiplayer_informations.STEPTREE
            if not len(multiplayer_info):
                multiplayer_info.append()

            # make sure there is falling damage info.
            falling_damages = meta.falling_damages.STEPTREE
            if not len(falling_damages):
                falling_damages.append()

        elif tag_cls == "metr":
            # The meter bitmaps can literally point to not
            # only the wrong tag, but the wrong TYPE of tag.
            # Since dependencies in meter tags are useless, we null them out.
            meta.stencil_bitmap.filepath = meta.source_bitmap.filepath = ''

        elif tag_cls in ("mode", "mod2"):
            if engine in ("halo1yelo", "halo1ce", "halo1pc", "halo1vap",
                          "halo1anni", "halo1pcdemo", "stubbspc"):
                # model_magic seems to be the same for all pc maps
                verts_start = tag_index.model_data_offset
                tris_start  = verts_start + tag_index.vertex_data_size
                model_magic = None
            else:
                model_magic = magic

            if model_magic is None:
                verts_attr_name = "uncompressed_vertices"
                byteswap_verts = byteswap_uncomp_verts
                vert_size = 68

                if engine != "stubbspc":
                    # need to swap the lod cutoff and nodes values around
                    cutoffs = (meta.superlow_lod_cutoff, meta.low_lod_cutoff,
                               meta.high_lod_cutoff, meta.superhigh_lod_cutoff)
                    meta.superlow_lod_cutoff  = cutoffs[3]
                    meta.low_lod_cutoff       = cutoffs[2]
                    meta.high_lod_cutoff      = cutoffs[1]
                    meta.superhigh_lod_cutoff = cutoffs[0]

            else:
                verts_attr_name = "compressed_vertices"
                byteswap_verts = byteswap_comp_verts
                vert_size = 32

            # grab vertices and indices from the map
            for geom in meta.geometries.STEPTREE:
                for part in geom.parts.STEPTREE:
                    verts_block = part[verts_attr_name]
                    tris_block  = part.triangles
                    info  = part.model_meta_info

                    # null out certain things in the part
                    part.previous_part_index = part.next_part_index = 0
                    part.centroid_primary_node = 0
                    part.centroid_secondary_node = 0
                    part.centroid_primary_weight = 0.0
                    part.centroid_secondary_weight = 0.0

                    # make the new blocks to hold the raw data
                    verts_block.STEPTREE = raw_block_def.build()
                    tris_block.STEPTREE  = raw_block_def.build()

                    # read the offsets of the vertices and indices from the map
                    if engine == "stubbspc":
                        verts_off = verts_start + info.vertices_reflexive_offset
                        tris_off  = tris_start  + info.indices_reflexive_offset
                    elif model_magic is None:
                        verts_off = verts_start + info.vertices_offset
                        tris_off  = tris_start  + info.indices_offset
                    else:
                        map_data.seek(
                            info.vertices_reflexive_offset + 4 - model_magic)
                        verts_off = unpack(
                            "<I", map_data.read(4))[0] - model_magic
                        map_data.seek(
                            info.indices_reflexive_offset  + 4 - model_magic)
                        tris_off  = unpack(
                            "<I", map_data.read(4))[0] - model_magic

                    # read the raw data from the map
                    map_data.seek(verts_off)
                    raw_verts = map_data.read(vert_size*info.vertex_count)
                    map_data.seek(tris_off)
                    raw_tris  = map_data.read(2*(info.index_count + 2))

                    # append the padding indices to the triangle strip
                    if len(raw_tris)%6 == 2:
                        raw_tris += b'\xff\xff\xff\xff'
                    elif len(raw_tris)%6 == 4:
                        raw_tris += b'\xff\xff'

                    # put the raw data in the verts and tris blocks
                    verts_block.STEPTREE.data = raw_verts
                    tris_block.STEPTREE.data  = raw_tris

                    # call the byteswappers
                    if byteswap:
                        byteswap_verts(verts_block)
                        byteswap_tris(tris_block)

                    # set the size of the reflexives
                    # this is NOT redundant. anniversary maps wont byteswap,
                    # so the size setting occuring in the byteswap wont happen.
                    verts_block.size = len(verts_block.STEPTREE.data) // vert_size
                    tris_block.size  = len(tris_block.STEPTREE.data) // 6

                    # null out the model_meta_info
                    for i in range(len(info)):
                        if isinstance(info[i], int):
                            info[i] = 0

        elif tag_cls == "pphy":
            # set the meta-only values to 0
            meta.scaled_density = 0
            meta.water_gravity_scale = 0
            meta.air_gravity_scale = 0

            # scale friction values
            meta.air_friction /= 10000
            meta.water_friction /= 10000

        elif tag_cls == "proj":
            # need to scale velocities by 30
            meta.proj_attrs.physics.initial_velocity *= 30
            meta.proj_attrs.physics.final_velocity *= 30

        elif tag_cls == "sbsp":
            if byteswap:
                byteswap_sbsp_meta(meta)

            # null out the runtime decals
            del meta.runtime_decals.STEPTREE[:]

            for cluster in meta.clusters.STEPTREE:
                predicted_resources.append(cluster.predicted_resources)

            compressed = "xbox" in engine or engine in ("stubbs", "shadowrun_proto")

            if compressed:
                generate_verts = kwargs.get("generate_uncomp_verts", False)
            else:
                generate_verts = kwargs.get("generate_comp_verts", False)

            endian = "<"
            if engine == "halo1anni":
                endian = ">"

            comp_norm   = compress_normal32
            decomp_norm = decompress_normal32

            comp_vert_nbt_unpacker = MethodType(unpack, endian + "3I")
            uncomp_vert_nbt_packer = MethodType(pack_into, endian + "12s9f8s")

            comp_vert_nuv_unpacker = MethodType(unpack, endian + "I2h")
            uncomp_vert_nuv_packer = MethodType(pack_into, endian + "5f")

            uncomp_vert_nbt_unpacker = MethodType(unpack, endian + "9f")
            comp_vert_nbt_packer = MethodType(pack_into, endian + "12s3I8s")

            uncomp_vert_nuv_unpacker = MethodType(unpack, endian + "5f")
            comp_vert_nuv_packer = MethodType(pack_into, endian + "I2h")

            for lightmap in meta.lightmaps.STEPTREE:
                for b in lightmap.materials.STEPTREE:
                    # need to null these or switching bsps will crash sapien
                    b.unknown_meta_offset0 = b.unknown_meta_offset1 = 0
                    b.vertices_meta_offset = 0
                    b.lightmap_vertices_meta_offset = 0
                    b.vertex_type.data = 0

                    if not generate_verts:
                        continue

                    vert_count = b.vertices_count
                    lightmap_vert_count = b.lightmap_vertices_count

                    u_verts = b.uncompressed_vertices
                    c_verts = b.compressed_vertices

                    if compressed:
                        # generate uncompressed vertices from the compressed
                        comp_buffer   = c_verts.STEPTREE
                        uncomp_buffer = bytearray(56*vert_count +
                                                  20*lightmap_vert_count)
                        in_off  = 0
                        out_off = 0
                        for i in range(vert_count):
                            n, b, t = comp_vert_nbt_unpacker(
                                comp_buffer[in_off + 12: in_off + 24])

                            # write the uncompressed data
                            uncomp_vert_nbt_packer(
                                uncomp_buffer, out_off,
                                comp_buffer[in_off: in_off + 12],
                                *decomp_norm(n),
                                *decomp_norm(b),
                                *decomp_norm(t),
                                comp_buffer[in_off + 24: in_off + 32])

                            in_off  += 32
                            out_off += 56

                        for i in range(lightmap_vert_count):
                            n, u, v = comp_vert_nuv_unpacker(
                                comp_buffer[in_off: in_off + 8])
                            # write the uncompressed data
                            uncomp_vert_nuv_packer(
                                uncomp_buffer, out_off,
                                *decomp_norm(n), u/32767, v/32767)

                            in_off  += 8
                            out_off += 20
                    else:
                        # generate compressed vertices from uncompressed
                        uncomp_buffer = u_verts.STEPTREE
                        comp_buffer   = bytearray(32*vert_count +
                                                  8*lightmap_vert_count)

                        in_off  = 0
                        out_off = 0
                        # for speed purposes, we'll assume all vectors
                        # are already normalized to a length of ~1.0
                        for i in range(vert_count):
                            ni, nj, nk, bi, bj, bk, ti, tj, tk = \
                                uncomp_vert_nbt_unpacker(
                                    uncomp_buffer[in_off + 12: in_off + 48])

                            # write the compressed data
                            comp_vert_nbt_packer(
                                comp_buffer, out_off,
                                uncomp_buffer[in_off: in_off + 12],
                                comp_norm(ni, nj, nk),
                                comp_norm(bi, bj, bk),
                                comp_norm(ti, tj, tk),
                                uncomp_buffer[in_off + 48: in_off + 56])

                            in_off  += 56
                            out_off += 32

                        for i in range(lightmap_vert_count):
                            ni, nj, nk, u, v = uncomp_vert_nuv_unpacker(
                                uncomp_buffer[in_off: in_off + 20])

                            # write the compressed data
                            comp_vert_nuv_packer(
                                comp_buffer, out_off,
                                comp_norm(ni, nj, nk),
                                int(min(max(u, -1.0), 1.0)*32767),
                                int(min(max(v, -1.0), 1.0)*32767))

                            in_off  += 20
                            out_off += 8

                    # replace the buffers
                    u_verts.STEPTREE = uncomp_buffer
                    c_verts.STEPTREE = comp_buffer

        elif tag_cls == "scnr":
            # need to remove the references to the child scenarios
            del meta.child_scenarios.STEPTREE[:]

            # set the bsp pointers and stuff to 0
            for b in meta.structure_bsps.STEPTREE:
                b.bsp_pointer = b.bsp_size = b.bsp_magic = 0

            predicted_resources.append(meta.predicted_resources)

            # byteswap the script syntax data
            if byteswap:
                byteswap_scnr_script_syntax_data(meta)

            # rename duplicate stuff that causes errors when compiling scripts
            if kwargs.get("rename_scnr_dups", False):
                string_data = meta.script_string_data.data.decode("latin-1")
                syntax_data = get_hsc_data_block(raw_syntax_data=meta.script_syntax_data.data)

                # NOTE: For a list of all the script object types
                # with their corrosponding enum value, check
                #     reclaimer.enums.script_object_types
                keep_these = {i: set() for i in
                              SCRIPT_OBJECT_TYPES_TO_SCENARIO_REFLEXIVES}
                for b in meta.bsp_switch_trigger_volumes.STEPTREE:
                    keep_these[11].add(b.trigger_volume)

                for i in range(min(syntax_data.last_node, len(syntax_data.nodes))):
                    node = syntax_data.nodes[i]
                    if node.type not in keep_these:
                        continue

                    keep_these[node.type].add(node.data & 0xFFff)

                for script_object_type, reflexive_name in \
                        SCRIPT_OBJECT_TYPES_TO_SCENARIO_REFLEXIVES.items():
                    keep = keep_these[script_object_type]
                    reflexive = meta[reflexive_name].STEPTREE
                    counts = {b.name.lower(): 0 for b in reflexive}
                    for b in reflexive:
                        counts[b.name.lower()] += 1

                    for i in range(len(reflexive)):
                        name = reflexive[i].name.lower()
                        if counts[name] > 1 and i not in keep:
                            reflexive[i].name = ("DUP%s~%s" % (i, name))[: 31]

            # divide the cutscene times by 30(they're in ticks) and
            # subtract the fade-in time from the up_time(normally added
            # together as a total up-time in maps, but not in tag form)
            for b in meta.cutscene_titles.STEPTREE:
                b.up_time = max(b.up_time - b.fade_in_time, 0.0)

                b.fade_in_time  /= 30
                b.fade_out_time /= 30
                b.up_time       /= 30

        elif tag_cls == "snd!":
            meta.maximum_bend_per_second = meta.maximum_bend_per_second ** 30
            for pitch_range in meta.pitch_ranges.STEPTREE:
                if not byteswap: break
                for permutation in pitch_range.permutations.STEPTREE:
                    if permutation.compression.enum_name == "none":
                        # byteswap pcm audio
                        byteswap_pcm16_samples(permutation.samples)

        elif tag_cls == "shpp":
            predicted_resources.append(meta.predicted_resources)

        elif tag_cls == "shpg":
            shpg_attrs = meta.shpg_attrs

            # copy all merged values into their respective reflexives
            for b in shpg_attrs.merged_values.STEPTREE:
                typ = b.value_type.enum_name
                cnt = b.value_count
                if   typ == "boolean": array = shpg_attrs.booleans.STEPTREE
                elif typ == "integer": array = shpg_attrs.integers.STEPTREE
                elif typ == "color":   array = shpg_attrs.colors.STEPTREE
                elif typ == "bitmap":  array = shpg_attrs.bitmaps.STEPTREE
                elif typ != "float":   continue  # unknown type
                elif cnt == 1: array = shpg_attrs.floats_1d.STEPTREE
                elif cnt == 2: array = shpg_attrs.floats_2d.STEPTREE
                elif cnt == 3: array = shpg_attrs.floats_3d.STEPTREE
                elif cnt == 4: array = shpg_attrs.floats_4d.STEPTREE
                else:          continue  # unknown float type

                array.append()
                new_b = array[-1]
                new_b.value_name = b.value_name
                values = b.values.u_node

                if typ == "bitmap":
                    new_b.bitmap = b.bitmap
                    new_b.bitmap_index = values.bitmap_index
                    continue

                new_b.runtime_value      = b.runtime_value
                new_b.animation_function = b.animation_function
                new_b.animation_flags    = b.animation_flags
                new_b.animation_duration = b.animation_duration
                new_b.animation_rate     = b.animation_rate

                if typ == "boolean":
                    new_b.flags = b.flags
                    new_b.value = values.value
                else:
                    new_b.value_lower_bound = values.value_lower_bound
                    new_b.value_upper_bound = values.value_upper_bound

            # clear the merged values reflexive
            del shpg_attrs.merged_values.STEPTREE[:]

        elif tag_cls == "weap":
            predicted_resources.append(meta.weap_attrs.predicted_resources)

        # remove any predicted resources
        for pr in predicted_resources:
            del pr.STEPTREE[:]

        return meta

    def get_resource_map_paths(self, maps_dir=""):
        if self.is_resource or self.engine not in ("halo1pc", "halo1pcdemo",
                                                   "halo1ce", "halo1yelo",
                                                   "halo1vap"):
            return {}

        map_paths = {"bitmaps": None, "sounds": None, "loc": None}
        if self.engine not in ("halo1ce", "halo1yelo", "halo1vap"):
            map_paths.pop('loc')

        data_files = False
        if hasattr(self.map_header, "yelo_header"):
            data_files = self.map_header.yelo_header.flags.uses_mod_data_files

        if not is_path_empty(maps_dir):
            maps_dir = Path(maps_dir)
        elif data_files:
            maps_dir = self.filepath.parent.joinpath("data_files")
        else:
            maps_dir = self.filepath.parent

        map_name_str = "%s.map"
        if data_files:
            map_name_str = "~" + map_name_str

        # detect the map paths for the resource maps
        for map_name in sorted(map_paths.keys()):
            map_path = maps_dir.joinpath(map_name_str % map_name)
            if self.maps.get(map_name) is not None:
                map_paths[map_name] = self.maps[map_name].filepath
            elif map_path.is_file():
                map_paths[map_name] = map_path

        return map_paths

    def generate_map_info_string(self):
        string = HaloMap.generate_map_info_string(self)
        index, header = self.tag_index, self.map_header

        string += """

Calculated information:
    index magic == %s
    map magic   == %s

Tag index:
    tag count           == %s
    scenario tag id     == %s
    index array pointer == %s   non-magic == %s
    model data pointer  == %s
    meta data length    == %s
    vertex parts count  == %s
    index  parts count  == %s""" % (
        self.index_magic, self.map_magic,
        index.tag_count, index.scenario_tag_id & 0xFFff,
        index.tag_index_offset, index.tag_index_offset - self.map_magic,
        index.model_data_offset, header.tag_data_size,
        index.vertex_parts_count, index.index_parts_count)

        if index.SIZE == 36:
            string += """
    index parts pointer == %s   non-magic == %s""" % (
        index.index_parts_offset, index.index_parts_offset - self.map_magic)
        else:
            string += """
    vertex data size    == %s
    index  data size    == %s
    model  data size    == %s""" % (
        index.vertex_data_size,
        index.model_data_size - index.vertex_data_size,
        index.model_data_size)

        string += "\n\nSbsp magic and headers:\n"
        for tag_id in self.bsp_magics:
            header = self.bsp_headers.get(tag_id)
            if header is None: continue

            magic  = self.bsp_magics[tag_id]
            string += """    %s.structure_scenario_bsp
        bsp base pointer     == %s
        bsp magic            == %s
        bsp size             == %s
        bsp metadata pointer == %s   non-magic == %s\n""" % (
            index.tag_index[tag_id].path, self.bsp_header_offsets[tag_id],
            magic, self.bsp_sizes[tag_id], header.meta_pointer,
            header.meta_pointer - magic)

        if self.engine == "halo1yelo":
            string += self.generate_yelo_info_string()
        elif self.engine == "halo1vap":
            string += self.generate_vap_info_string()

        return string

    def generate_yelo_info_string(self):
        yelo    = self.map_header.yelo_header
        flags   = yelo.flags
        info    = yelo.build_info
        version = yelo.tag_versioning
        cheape  = yelo.cheape_definitions
        rsrc    = yelo.resources
        min_os  = info.minimum_os_build

        return """
Yelo information:
    Mod name              == %s
    Memory upgrade amount == %sx

    Flags:
        uses memory upgrades       == %s
        uses mod data files        == %s
        is protected               == %s
        uses game state upgrades   == %s
        has compression parameters == %s

    Build info:
        build string  == %s
        timestamp     == %s
        stage         == %s
        revision      == %s

    Cheape:
        build string      == %s
        version           == %s.%s.%s
        size              == %s
        offset            == %s
        decompressed size == %s

    Versioning:
        minimum open sauce     == %s.%s.%s
        project yellow         == %s
        project yellow globals == %s

    Resources:
        compression parameters header offset   == %s
        tag symbol storage header offset       == %s
        string id storage header offset        == %s
        tag string to id storage header offset == %s\n""" % (
            yelo.mod_name, yelo.memory_upgrade_multiplier,
            bool(flags.uses_memory_upgrades),
            bool(flags.uses_mod_data_files),
            bool(flags.is_protected),
            bool(flags.uses_game_state_upgrades),
            bool(flags.has_compression_params),
            info.build_string, info.timestamp, info.stage.enum_name,
            info.revision, cheape.build_string,
            info.cheape.maj, info.cheape.min, info.cheape.build,
            cheape.size, cheape.offset, cheape.decompressed_size,
            min_os.maj, min_os.min, min_os.build,
            version.project_yellow, version.project_yellow_globals,
            rsrc.compression_params_header_offset,
            rsrc.tag_symbol_storage_header_offset,
            rsrc.string_id_storage_header_offset,
            rsrc.tag_string_to_id_storage_header_offset)

    def generate_vap_info_string(self):
        vap = self.map_header.vap_header

        return """
VAP information:
    name        == %s
    build date  == %s
    description == %s

    vap version   == %s
    feature level == %s
    max players   == %s\n""" % (
            vap.name, vap.build_date, vap.description,
            vap.vap_version.enum_name, vap.feature_level.enum_name,
            vap.max_players,
            )



#
# This file is part of Reclaimer.
#
# For authors and copyright check AUTHORS.TXT
#
# Reclaimer is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import traceback

from copy import deepcopy

from reclaimer import enums
from reclaimer.animation import util
from reclaimer.animation import serialization

__all__ = ("compile_animation", "compile_model_animations",
           "ANIMATION_COMPILE_MODE_NEW", "ANIMATION_COMPILE_MODE_PRESERVE",
           "ANIMATION_COMPILE_MODE_ADDITIVE")

ANIMATION_COMPILE_MODE_NEW = 0
ANIMATION_COMPILE_MODE_PRESERVE = 1
ANIMATION_COMPILE_MODE_ADDITIVE = 2


def compile_animation(anim, jma_anim, ignore_size_limits=False, endian=">"):
    '''
    Compiles the provided JmaAnimation into the provided antr animation block.
    '''
    errors = []

    # determine the sizes of the frame_info, default_data, and frame_data
    frame_info_node_size = jma_anim.root_node_info_frame_size

    stored_frame_count = jma_anim.frame_count - 1  # subtract the base frame

    frame_info_size   = frame_info_node_size * stored_frame_count
    default_data_size = jma_anim.default_data_size
    frame_data_size   = jma_anim.frame_data_frame_size * stored_frame_count

    max_frame_info_size   = anim.frame_info.get_desc('MAX', 'size')
    max_default_data_size = anim.default_data.get_desc('MAX', 'size')
    max_frame_data_size   = anim.frame_data.get_desc('MAX', 'size')

    if not ignore_size_limits:
        if frame_info_size > max_frame_info_size:
            errors.append("Too much frame_info data. Max is %s bytes, got %s" %
                          (max_frame_info_size, frame_info_size))

        if default_data_size > max_default_data_size:
            errors.append("Too much default data. Max is %s bytes, got %s" %
                          (max_default_data_size, default_data_size))

        if frame_data_size > max_frame_data_size:
            errors.append("Too much frame data. Max is %s bytes, got %s" %
                          (max_frame_data_size, frame_data_size))

    if errors:
        return errors

    if jma_anim.root_node_info_applied:
        jma_anim = deepcopy(jma_anim)
        jma_anim.apply_root_node_info_to_states(True)

    anim.name = jma_anim.name
    anim.type.set_to(jma_anim.anim_type)
    anim.frame_count = stored_frame_count
    anim.frame_size = jma_anim.frame_data_frame_size
    anim.node_list_checksum = jma_anim.node_list_checksum
    anim.node_count = jma_anim.node_count
    anim.next_animation = -1

    if jma_anim.has_dz:
        anim.frame_info_type.data = 3
    elif jma_anim.has_dyaw:
        anim.frame_info_type.data = 2
    elif jma_anim.has_dxdy:
        anim.frame_info_type.data = 1
    else:
        anim.frame_info_type.data = 0

    # cant compress animations yet
    anim.flags.data = 0
    anim.flags.world_relative = jma_anim.world_relative
    anim.offset_to_compressed_data = 0

    anim.trans_flags0 =  jma_anim.trans_flags_int & 0xFFffFFff
    anim.trans_flags1 = (jma_anim.trans_flags_int >> 32) & 0xFFffFFff
    anim.rot_flags0   =  jma_anim.rot_flags_int & 0xFFffFFff
    anim.rot_flags1   = (jma_anim.rot_flags_int >> 32) & 0xFFffFFff
    anim.scale_flags0 =  jma_anim.scale_flags_int & 0xFFffFFff
    anim.scale_flags1 = (jma_anim.scale_flags_int >> 32) & 0xFFffFFff

    frame_info = serialization.serialize_frame_info(jma_anim, endian)
    def_data   = serialization.serialize_default_data(jma_anim, endian)
    frame_data = serialization.serialize_frame_data(jma_anim, endian)

    anim.frame_info.STEPTREE = frame_info
    anim.default_data.STEPTREE = def_data
    anim.frame_data.STEPTREE = frame_data

    return errors


def compile_model_animations(antr_tag, jma_anim_set, ignore_size_limits=False,
                             animation_count_limit=256, delta_tolerance=None,
                             update_mode=ANIMATION_COMPILE_MODE_PRESERVE,
                             mod2_nodes=None):
    errors = []

    tagdata = antr_tag.data.tagdata
    antr_objects = tagdata.objects.STEPTREE
    antr_units = tagdata.units.STEPTREE
    antr_weapons = tagdata.weapons.STEPTREE
    antr_vehicles = tagdata.vehicles.STEPTREE
    antr_devices = tagdata.devices.STEPTREE
    antr_unit_damages = tagdata.unit_damages.STEPTREE
    antr_fp_animations = tagdata.fp_animations.STEPTREE
    antr_anims = tagdata.animations.STEPTREE
    antr_nodes = tagdata.nodes.STEPTREE

    # make sure the new animations share nodes with the existing animations
    if update_mode == ANIMATION_COMPILE_MODE_ADDITIVE and antr_nodes:
        if len(antr_nodes) != len(jma_anim_set.nodes):
            errors.append("Node count of these animations differs from the "
                          "node count of the model_animations tag.")
            return errors

        for i in range(len(jma_anim_set.nodes)):
            jma_node = jma_anim_set.nodes[i]
            antr_node = antr_nodes[i]

            if (antr_node.next_sibling_node_index != jma_node.sibling_index or
                antr_node.first_child_node_index  != jma_node.first_child or
                # dont check the parent node. root node references itself in
                # animation nodes. it's weird. Sibling and child are enough.
                #antr_node.parent_node_index       != jma_node.parent_index or
                antr_node.name != jma_node.name):

                errors.append("Node(s) in these animations differ from the "
                              "node(s) of the model_animations tag.")
                return errors


    prev_antr_anims_by_type_strings = {}
    antr_indices_by_type_strings = {}

    prev_antr_objects  = list(antr_objects)
    prev_antr_units    = list(antr_units)
    prev_antr_vehicles = list(antr_vehicles)
    prev_antr_nodes  = list(antr_nodes)
    prev_antr_anims  = list(antr_anims)
    assert update_mode in (ANIMATION_COMPILE_MODE_NEW,
                           ANIMATION_COMPILE_MODE_PRESERVE,
                           ANIMATION_COMPILE_MODE_ADDITIVE)
    if update_mode != ANIMATION_COMPILE_MODE_ADDITIVE:
        if update_mode == ANIMATION_COMPILE_MODE_NEW:
            del tagdata.sound_references.STEPTREE[:]
            tagdata.limp_body_node_radius = 0.0
            tagdata.flags.data = 0
            del prev_antr_objects[:]
            del prev_antr_units[:]
            del prev_antr_vehicles[:]
            del prev_antr_nodes[:]
            del prev_antr_anims[:]

        del antr_objects[:]
        del antr_units[:]
        del antr_weapons[:]
        del antr_vehicles[:]
        del antr_devices[:]
        del antr_unit_damages[:]
        del antr_fp_animations[:]
        del antr_anims[:]
        del antr_nodes[:]

        for i in range(len(jma_anim_set.nodes)):
            jma_node = jma_anim_set.nodes[i]
            antr_nodes.append()
            antr_node = antr_nodes[-1]
            antr_node.name = jma_node.name
            antr_node.next_sibling_node_index = jma_node.sibling_index
            antr_node.first_child_node_index  = jma_node.first_child
            antr_node.parent_node_index       = jma_node.parent_index
            if antr_node.parent_node_index == -1:
                antr_node.parent_node_index = 0

            if i not in range(len(prev_antr_nodes)):
                continue

        if prev_antr_nodes:
            prev_antr_node = prev_antr_nodes[i]
            if prev_antr_node.name.lower() == antr_node.name.lower():
                antr_node.node_joint_flags.data = prev_antr_node.node_joint_flags.data
                antr_node.base_vector[:] = prev_antr_node.base_vector
                antr_node.vector_range   = prev_antr_node.vector_range

    # cache the old animations by their names
    for i in range(len(prev_antr_anims)):
        anim = prev_antr_anims[i]
        _, name_pieces = util.split_anim_name_into_type_strings(anim.name.strip())
        prev_antr_anims_by_type_strings[name_pieces] = anim

    # cache the existing animation indices by their names
    for i in range(len(antr_anims)):
        anim = antr_anims[i]
        _, name_pieces = util.split_anim_name_into_type_strings(anim.name.strip())
        antr_indices_by_type_strings[name_pieces] = i


    indices_to_not_overwrite = set()
    indices_modified = set()
    # loop over the animations to add and add/replace them
    for jma_anim_name in sorted(jma_anim_set.animations):
        name = jma_anim_name.strip()
        has_purpose, name_pieces = util.split_anim_name_into_type_strings(name)

        # find where to put this animation
        anim_index = antr_indices_by_type_strings.get(
            name_pieces, len(antr_anims))

        if anim_index >= animation_count_limit:
            errors.append(
                "Too many animations. Cannot add '%s'" % jma_anim_name)
            continue

        anim_added = False
        if anim_index == len(antr_anims):
            anim_added = True
            antr_anims.append()
            print("Adding '%s'" % jma_anim_name)
        else:
            print("Replacing '%s'" % jma_anim_name)

        anim = antr_anims[anim_index]
        try:
            jma_anim = jma_anim_set.animations[jma_anim_name]
            jma_anim.calculate_animation_flags(delta_tolerance)
            errors.extend(compile_animation(anim, jma_anim, ignore_size_limits))
        except Exception:
            errors.append(traceback.format_exc())
            errors.append("Could not compile '%s'" % jma_anim_name)
            if anim_added:
                antr_anims.pop(anim_index)

            continue

        # update the antr indices with this animations index
        antr_indices_by_type_strings[name_pieces] = anim_index

        if name_pieces in prev_antr_anims_by_type_strings:
            # update the animation with the old ones keyframe values and such
            prev_anim = prev_antr_anims_by_type_strings[name_pieces]
            anim.weight = prev_anim.weight
            anim.sound  = prev_anim.sound
            anim.loop_frame_index       = prev_anim.loop_frame_index
            anim.key_frame_index        = prev_anim.key_frame_index
            anim.second_key_frame_index = prev_anim.second_key_frame_index
            anim.sound_frame_index      = prev_anim.sound_frame_index
            anim.left_foot_frame_index  = prev_anim.left_foot_frame_index
            anim.right_foot_frame_index = prev_anim.right_foot_frame_index


        if not has_purpose:
            print("    Could not determine a purpose for this animation")
            continue

        try:
            if util.set_animation_index(antr_tag, jma_anim_name, anim_index,
                                        indices_to_not_overwrite):
                # successfully found an animation index to use this animation
                indices_to_not_overwrite.add(anim_index)
            indices_modified.add(anim_index)
        except Exception:
            errors.append(traceback.format_exc())

    if update_mode == ANIMATION_COMPILE_MODE_PRESERVE:
        for i in range(len(prev_antr_objects)):
            prev_obje = prev_antr_objects[i]
            if prev_obje.animation not in range(len(prev_antr_anims)):
                continue

            _, name_pieces = util.split_anim_name_into_type_strings(
                prev_antr_anims[prev_obje.animation].name.strip())

            new_anim_index = antr_indices_by_type_strings.get(name_pieces)
            if new_anim_index is None:
                continue

            antr_objects.append()
            obje = antr_objects[-1]

            obje.animation = new_anim_index
            obje.function.data = prev_obje.function.data
            obje.function_controls.data = prev_obje.function_controls.data

        if antr_units and prev_antr_units:
            # preserve any old unit pitch/yaw/weapon values
            copy_unit_animation_block_data(antr_units, prev_antr_units)

        if antr_vehicles and prev_antr_vehicles:
            # preserve any old vehicle pitch/yaw/suspension values
            copy_vehicle_animation_block_data(antr_vehicles, prev_antr_vehicles)


    if (update_mode != ANIMATION_COMPILE_MODE_ADDITIVE and
        mod2_nodes and antr_units and not antr_vehicles):
        # only calculate node vectors for units that are not vehicles
        if len(antr_nodes) != len(mod2_nodes):
            errors.append("Gbxmodel node count does not match node "
                          "count in the model_animations tag.")
            return errors
        util.calculate_node_vectors(antr_nodes, mod2_nodes,
                                    jma_anim_set.animations.values())

    # fill in the remaining unit damages
    if antr_unit_damages:
        if len(antr_unit_damages) < 176:
            antr_unit_damages.extend(176 - len(antr_unit_damages))

        # loop over  s-ping, h-ping, s-kill, h-kill
        for i in range(0, len(antr_unit_damages), 44):
            # loop over each of the 11 animations per set
            for j in range(11):
                # make a collection of defaults for all sides of this region
                defaults = {i + j + k: -1 for k in (0, 11, 22, 33)}
                util.get_default_animation_enums(antr_unit_damages, defaults)

                # default any unset sides for this region
                util.set_default_animation_enums(antr_unit_damages, defaults)


    unit_anim_defaults = {enums.unit_animation_names.index(name): -1
                          for name in util.SHARED_UNIT_ANIMATION_NAMES}
    unit_weap_anim_defaults = {enums.unit_weapon_animation_names.index(name): -1
                               for name in util.SHARED_UNIT_WEAPON_ANIMATION_NAMES}

    # get the unit animations with applicable ones from all units.
    # do this in reverse since thats what tool seems to do.
    for unit in antr_units[::-1]:
        anim_enums = unit.animations.STEPTREE
        util.get_default_animation_enums(anim_enums, unit_anim_defaults)
        for unit_weap in unit.weapons.STEPTREE:
            anim_enums = unit_weap.animations.STEPTREE
            util.get_default_animation_enums(anim_enums, unit_weap_anim_defaults)

    # strip any unused animation indices
    unit_anim_defaults = {
        k: v for k, v in unit_anim_defaults.items() if v != -1}
    unit_weap_anim_defaults = {
        k: v for k, v in unit_weap_anim_defaults.items() if v != -1}

    for unit in antr_units:
        if unit_anim_defaults:
            anim_enums = unit.animations.STEPTREE
            anim_enums.extend(max(unit_anim_defaults) + 1 - len(anim_enums))

        if unit_weap_anim_defaults:
            for unit_weap in unit.weapons.STEPTREE:
                anim_enums = unit_weap.animations.STEPTREE
                anim_enums.extend(max(unit_weap_anim_defaults) + 1 - len(anim_enums))

    # default any unset unit animations with the found defaults.
    for unit in antr_units:
        util.set_default_animation_enums(unit.animations.STEPTREE,
                                         unit_anim_defaults)
        for unit_weap in unit.weapons.STEPTREE:
            util.set_default_animation_enums(unit_weap.animations.STEPTREE,
                                             unit_weap_anim_defaults)

    # setup permutation indices
    last_perm_anim_index = -1
    last_perm_name_pieces = ()
    # loop over the name pieces in a sorted manner so the lowest
    # permutations come first and are followed by the rest of them
    for name_pieces in sorted(antr_indices_by_type_strings):
        anim_index = antr_indices_by_type_strings[name_pieces]
        name_pieces, perm_num = name_pieces[: -1], name_pieces[-1]
        if (anim_index in indices_modified and
            name_pieces == last_perm_name_pieces and
            last_perm_anim_index != -1):
            antr_anims[last_perm_anim_index].next_animation = anim_index

        last_perm_anim_index = anim_index
        last_perm_name_pieces = name_pieces

    return errors


def copy_unit_animation_block_data(antr_units, prev_antr_units):
    for unit in antr_units:
        prev_unit = None
        # find the old unit with a label matching this one
        for block in prev_antr_units:
            if block.label == unit.label:
                prev_unit = block
                break

        if prev_unit is None: continue
        unit.right_yaw_per_frame  = prev_unit.right_yaw_per_frame
        unit.left_yaw_per_frame   = prev_unit.left_yaw_per_frame
        unit.right_frame_count    = prev_unit.right_frame_count
        unit.left_frame_count     = prev_unit.left_frame_count

        unit.down_pitch_per_frame = prev_unit.down_pitch_per_frame
        unit.up_pitch_per_frame   = prev_unit.up_pitch_per_frame
        unit.down_frame_count     = prev_unit.down_frame_count
        unit.up_frame_count       = prev_unit.up_frame_count

        unit.ik_points = deepcopy(prev_unit.ik_points)

        for unit_weap in unit.weapons.STEPTREE:
            prev_unit_weap = None
            # find the old unit with a label matching this one
            for block in prev_unit.weapons.STEPTREE:
                if block.name == unit_weap.name:
                    prev_unit_weap = block
                    break

            if prev_unit_weap is None: continue
            unit_weap.grip_marker  = prev_unit_weap.grip_marker
            unit_weap.hand_marker  = prev_unit_weap.hand_marker

            unit_weap.right_yaw_per_frame  = prev_unit_weap.right_yaw_per_frame
            unit_weap.left_yaw_per_frame   = prev_unit_weap.left_yaw_per_frame
            unit_weap.right_frame_count    = prev_unit_weap.right_frame_count
            unit_weap.left_frame_count     = prev_unit_weap.left_frame_count

            unit_weap.down_pitch_per_frame = prev_unit_weap.down_pitch_per_frame
            unit_weap.up_pitch_per_frame   = prev_unit_weap.up_pitch_per_frame
            unit_weap.down_frame_count     = prev_unit_weap.down_frame_count
            unit_weap.up_frame_count       = prev_unit_weap.up_frame_count

            unit_weap.ik_points = deepcopy(prev_unit_weap.ik_points)


def copy_vehicle_animation_block_data(antr_vehicles, prev_antr_vehicles):
    vehicle = antr_vehicles[0]
    prev_vehicle = prev_antr_vehicles[0]
    vehicle.right_yaw_per_frame  = prev_vehicle.right_yaw_per_frame
    vehicle.left_yaw_per_frame   = prev_vehicle.left_yaw_per_frame
    vehicle.right_frame_count    = prev_vehicle.right_frame_count
    vehicle.left_frame_count     = prev_vehicle.left_frame_count

    vehicle.down_pitch_per_frame = prev_vehicle.down_pitch_per_frame
    vehicle.up_pitch_per_frame   = prev_vehicle.up_pitch_per_frame
    vehicle.down_frame_count     = prev_vehicle.down_frame_count
    vehicle.up_frame_count       = prev_vehicle.up_frame_count

    suspensions = vehicle.suspension_animations.STEPTREE
    prev_suspensions = prev_vehicle.suspension_animations.STEPTREE
    for prev, new in zip(prev_suspensions, suspensions):
        new.mass_point_index = prev.mass_point_index
        new.full_extension_ground_depth   = prev.full_extension_ground_depth
        new.full_compression_ground_depth = prev.full_compression_ground_depth
